from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_DLL, SDL_DLL

from .SDL_stdinc import SDL_FunctionPointer

SDL_SET_CURRENT_DLL(SDL_DLL)

class SDL_SharedObject(ctypes.c_void_p):
    ...

SDL_FUNC("SDL_LoadObject", ctypes.POINTER(SDL_SharedObject), ctypes.c_char_p)
SDL_FUNC("SDL_LoadFunction", SDL_FunctionPointer, ctypes.POINTER(SDL_SharedObject), ctypes.c_char_p)
SDL_FUNC("SDL_UnloadObject", None, ctypes.POINTER(SDL_SharedObject))


from .__init__ import sys, ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_DLL, SDL_DLL

from .SDL_init import SDL_AppResult, \
    SDL_AppInit_func, SDL_AppIterate_func, SDL_AppEvent_func, SDL_AppQuit_func

SDL_SET_CURRENT_DLL(SDL_DLL)

# SDL_FUNC("SDL_AppInit", SDL_AppResult, ctypes.POINTER(ctypes.c_void_p), ctypes.c_int, ctypes.POINTER(ctypes.c_char_p))
# SDL_FUNC("SDL_AppIterate", SDL_AppResult, ctypes.c_void_p)
# SDL_FUNC("SDL_AppEvent", SDL_AppResult, ctypes.c_void_p, ctypes.POINTER(SDL_Event))
# SDL_FUNC("SDL_AppQuit", None, ctypes.c_void_p, SDL_AppResult)

SDL_main_func = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_char_p))
# SDL_FUNC("SDL_main", ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_char_p))

SDL_FUNC("SDL_SetMainReady", None)

SDL_FUNC("SDL_RunApp", ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_char_p), SDL_main_func, ctypes.c_void_p)
SDL_FUNC("SDL_EnterAppMainCallbacks", ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_char_p), SDL_AppInit_func, SDL_AppIterate_func, SDL_AppEvent_func, SDL_AppQuit_func)

if "win32" in sys.platform:
    SDL_FUNC("SDL_RegisterApp", ctypes.c_bool, ctypes.c_char_p, ctypes.c_uint32, ctypes.c_void_p)
    SDL_FUNC("SDL_UnregisterApp", None)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_DLL, SDL_DLL

from .SDL_properties import SDL_PropertiesID

SDL_SET_CURRENT_DLL(SDL_DLL)

class SDL_Sensor(ctypes.c_void_p):
    ...

SDL_SensorID = ctypes.c_uint32

SDL_STANDARD_GRAVITY = 9.80665

SDL_SensorType = ctypes.c_int

SDL_SENSOR_INVALID = -1
SDL_SENSOR_UNKNOWN = 0
SDL_SENSOR_ACCEL = 1
SDL_SENSOR_GYRO = 2
SDL_SENSOR_ACCEL_L = 3
SDL_SENSOR_GYRO_L = 4
SDL_SENSOR_ACCEL_R = 5
SDL_SENSOR_GYRO_R = 6

SDL_FUNC("SDL_GetSensors", ctypes.POINTER(SDL_SensorID), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_GetSensorNameForID", ctypes.c_char_p, SDL_SensorID)
SDL_FUNC("SDL_GetSensorTypeForID", SDL_SensorType, SDL_SensorID)
SDL_FUNC("SDL_GetSensorNonPortableTypeForID", ctypes.c_int, SDL_SensorID)
SDL_FUNC("SDL_OpenSensor", ctypes.POINTER(SDL_Sensor), SDL_SensorID)
SDL_FUNC("SDL_GetSensorFromID", ctypes.POINTER(SDL_Sensor), SDL_SensorID)
SDL_FUNC("SDL_GetSensorProperties", SDL_PropertiesID, ctypes.POINTER(SDL_Sensor))
SDL_FUNC("SDL_GetSensorName", ctypes.c_char_p, ctypes.POINTER(SDL_Sensor))
SDL_FUNC("SDL_GetSensorType", SDL_SensorType, ctypes.POINTER(SDL_Sensor))
SDL_FUNC("SDL_GetSensorNonPortableType", ctypes.c_int, ctypes.POINTER(SDL_Sensor))
SDL_FUNC("SDL_GetSensorID", SDL_SensorID, ctypes.POINTER(SDL_Sensor))
SDL_FUNC("SDL_GetSensorData", ctypes.c_bool, ctypes.POINTER(SDL_Sensor), ctypes.POINTER(ctypes.c_float), ctypes.c_int)
SDL_FUNC("SDL_CloseSensor", None, ctypes.POINTER(SDL_Sensor))
SDL_FUNC("SDL_UpdateSensors", None)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_DLL, SDL_DLL

from .SDL_events import SDL_Event
from .SDL_rect import SDL_Point, SDL_FPoint, SDL_Rect, SDL_FRect
from .SDL_pixels import SDL_Color, SDL_FColor, SDL_PixelFormat
from .SDL_surface import SDL_Surface, SDL_ScaleMode, SDL_FlipMode
from .SDL_video import SDL_Window, SDL_WindowFlags
from .SDL_properties import SDL_PropertiesID
from .SDL_blendmode import SDL_BlendMode

SDL_SET_CURRENT_DLL(SDL_DLL)

SDL_SOFTWARE_RENDERER = "software"

class SDL_Vertex(ctypes.Structure):
    _fields_ = [
        ("position", SDL_FPoint),
        ("color", SDL_Color),
        ("tex_coord", SDL_FPoint)
    ]

SDL_TextureAccess = ctypes.c_int

SDL_TEXTUREACCESS_STATIC = 0
SDL_TEXTUREACCESS_STREAMING = 1
SDL_TEXTUREACCESS_TARGET = 2

SDL_RendererLogicalPresentation = ctypes.c_int

SDL_LOGICAL_PRESENTATION_DISABLED = 0
SDL_LOGICAL_PRESENTATION_STRETCH = 1
SDL_LOGICAL_PRESENTATION_LETTERBOX = 2
SDL_LOGICAL_PRESENTATION_OVERSCAN = 3
SDL_LOGICAL_PRESENTATION_INTEGER_SCALE = 4

class SDL_Renderer(ctypes.c_void_p):
    ...

class SDL_Texture(ctypes.c_void_p):
    ...

SDL_FUNC("SDL_GetNumRenderDrivers", ctypes.c_int)
SDL_FUNC("SDL_GetRenderDriver", ctypes.c_char_p, ctypes.c_int)

SDL_FUNC("SDL_CreateWindowAndRenderer", ctypes.c_bool, ctypes.c_char_p, ctypes.c_int, ctypes.c_int, SDL_WindowFlags, ctypes.POINTER(ctypes.POINTER(SDL_Window)), ctypes.POINTER(ctypes.POINTER(SDL_Renderer)))
SDL_FUNC("SDL_CreateRenderer", ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Window), ctypes.c_char_p)
SDL_FUNC("SDL_CreateRendererWithProperties", ctypes.POINTER(SDL_Renderer), SDL_PropertiesID)

SDL_PROP_RENDERER_CREATE_NAME_STRING = "SDL.renderer.create.name"
SDL_PROP_RENDERER_CREATE_WINDOW_POINTER = "SDL.renderer.create.window"
SDL_PROP_RENDERER_CREATE_SURFACE_POINTER = "SDL.renderer.create.surface"
SDL_PROP_RENDERER_CREATE_OUTPUT_COLORSPACE_NUMBER = "SDL.renderer.create.output_colorspace"
SDL_PROP_RENDERER_CREATE_PRESENT_VSYNC_NUMBER = "SDL.renderer.create.present_vsync"
SDL_PROP_RENDERER_CREATE_VULKAN_INSTANCE_POINTER = "SDL.renderer.create.vulkan.instance"
SDL_PROP_RENDERER_CREATE_VULKAN_SURFACE_NUMBER = "SDL.renderer.create.vulkan.surface"
SDL_PROP_RENDERER_CREATE_VULKAN_PHYSICAL_DEVICE_POINTER = "SDL.renderer.create.vulkan.physical_device"
SDL_PROP_RENDERER_CREATE_VULKAN_DEVICE_POINTER = "SDL.renderer.create.vulkan.device"
SDL_PROP_RENDERER_CREATE_VULKAN_GRAPHICS_QUEUE_FAMILY_INDEX_NUMBER = "SDL.renderer.create.vulkan.graphics_queue_family_index"
SDL_PROP_RENDERER_CREATE_VULKAN_PRESENT_QUEUE_FAMILY_INDEX_NUMBER = "SDL.renderer.create.vulkan.present_queue_family_index"

SDL_FUNC("SDL_CreateSoftwareRenderer", ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Surface))
SDL_FUNC("SDL_GetRenderer", ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Window))
SDL_FUNC("SDL_GetRenderWindow", ctypes.POINTER(SDL_Window), ctypes.POINTER(SDL_Renderer))
SDL_FUNC("SDL_GetRendererName", ctypes.c_char_p, ctypes.POINTER(SDL_Renderer))

SDL_FUNC("SDL_GetRendererProperties", SDL_PropertiesID, ctypes.POINTER(SDL_Renderer))

SDL_PROP_RENDERER_NAME_STRING = "SDL.renderer.name"
SDL_PROP_RENDERER_WINDOW_POINTER = "SDL.renderer.window"
SDL_PROP_RENDERER_SURFACE_POINTER = "SDL.renderer.surface"
SDL_PROP_RENDERER_VSYNC_NUMBER = "SDL.renderer.vsync"
SDL_PROP_RENDERER_MAX_TEXTURE_SIZE_NUMBER = "SDL.renderer.max_texture_size"
SDL_PROP_RENDERER_TEXTURE_FORMATS_POINTER = "SDL.renderer.texture_formats"
SDL_PROP_RENDERER_OUTPUT_COLORSPACE_NUMBER = "SDL.renderer.output_colorspace"
SDL_PROP_RENDERER_HDR_ENABLED_BOOLEAN = "SDL.renderer.HDR_enabled"
SDL_PROP_RENDERER_SDR_WHITE_POINT_FLOAT = "SDL.renderer.SDR_white_point"
SDL_PROP_RENDERER_HDR_HEADROOM_FLOAT = "SDL.renderer.HDR_headroom"
SDL_PROP_RENDERER_D3D9_DEVICE_POINTER = "SDL.renderer.d3d9.device"
SDL_PROP_RENDERER_D3D11_DEVICE_POINTER = "SDL.renderer.d3d11.device"
SDL_PROP_RENDERER_D3D11_SWAPCHAIN_POINTER = "SDL.renderer.d3d11.swap_chain"
SDL_PROP_RENDERER_D3D12_DEVICE_POINTER = "SDL.renderer.d3d12.device"
SDL_PROP_RENDERER_D3D12_SWAPCHAIN_POINTER = "SDL.renderer.d3d12.swap_chain"
SDL_PROP_RENDERER_D3D12_COMMAND_QUEUE_POINTER = "SDL.renderer.d3d12.command_queue"
SDL_PROP_RENDERER_VULKAN_INSTANCE_POINTER = "SDL.renderer.vulkan.instance"
SDL_PROP_RENDERER_VULKAN_SURFACE_NUMBER = "SDL.renderer.vulkan.surface"
SDL_PROP_RENDERER_VULKAN_PHYSICAL_DEVICE_POINTER = "SDL.renderer.vulkan.physical_device"
SDL_PROP_RENDERER_VULKAN_DEVICE_POINTER = "SDL.renderer.vulkan.device"
SDL_PROP_RENDERER_VULKAN_GRAPHICS_QUEUE_FAMILY_INDEX_NUMBER = "SDL.renderer.vulkan.graphics_queue_family_index"
SDL_PROP_RENDERER_VULKAN_PRESENT_QUEUE_FAMILY_INDEX_NUMBER = "SDL.renderer.vulkan.present_queue_family_index"
SDL_PROP_RENDERER_VULKAN_SWAPCHAIN_IMAGE_COUNT_NUMBER = "SDL.renderer.vulkan.swapchain_image_count"

SDL_FUNC("SDL_GetRenderOutputSize", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_GetCurrentRenderOutputSize", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))

SDL_FUNC("SDL_CreateTexture", ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_Renderer), SDL_PixelFormat, SDL_TextureAccess, ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_CreateTextureFromSurface", ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Surface))
SDL_FUNC("SDL_CreateTextureWithProperties", ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_Renderer), SDL_PropertiesID)

SDL_PROP_TEXTURE_CREATE_COLORSPACE_NUMBER = "SDL.texture.create.colorspace"
SDL_PROP_TEXTURE_CREATE_FORMAT_NUMBER = "SDL.texture.create.format"
SDL_PROP_TEXTURE_CREATE_ACCESS_NUMBER = "SDL.texture.create.access"
SDL_PROP_TEXTURE_CREATE_WIDTH_NUMBER = "SDL.texture.create.width"
SDL_PROP_TEXTURE_CREATE_HEIGHT_NUMBER = "SDL.texture.create.height"
SDL_PROP_TEXTURE_CREATE_SDR_WHITE_POINT_FLOAT = "SDL.texture.create.SDR_white_point"
SDL_PROP_TEXTURE_CREATE_HDR_HEADROOM_FLOAT = "SDL.texture.create.HDR_headroom"
SDL_PROP_TEXTURE_CREATE_D3D11_TEXTURE_POINTER = "SDL.texture.create.d3d11.texture"
SDL_PROP_TEXTURE_CREATE_D3D11_TEXTURE_U_POINTER = "SDL.texture.create.d3d11.texture_u"
SDL_PROP_TEXTURE_CREATE_D3D11_TEXTURE_V_POINTER = "SDL.texture.create.d3d11.texture_v"
SDL_PROP_TEXTURE_CREATE_D3D12_TEXTURE_POINTER = "SDL.texture.create.d3d12.texture"
SDL_PROP_TEXTURE_CREATE_D3D12_TEXTURE_U_POINTER = "SDL.texture.create.d3d12.texture_u"
SDL_PROP_TEXTURE_CREATE_D3D12_TEXTURE_V_POINTER = "SDL.texture.create.d3d12.texture_v"
SDL_PROP_TEXTURE_CREATE_METAL_PIXELBUFFER_POINTER = "SDL.texture.create.metal.pixelbuffer"
SDL_PROP_TEXTURE_CREATE_OPENGL_TEXTURE_NUMBER = "SDL.texture.create.opengl.texture"
SDL_PROP_TEXTURE_CREATE_OPENGL_TEXTURE_UV_NUMBER = "SDL.texture.create.opengl.texture_uv"
SDL_PROP_TEXTURE_CREATE_OPENGL_TEXTURE_U_NUMBER = "SDL.texture.create.opengl.texture_u"
SDL_PROP_TEXTURE_CREATE_OPENGL_TEXTURE_V_NUMBER = "SDL.texture.create.opengl.texture_v"
SDL_PROP_TEXTURE_CREATE_OPENGLES2_TEXTURE_NUMBER = "SDL.texture.create.opengles2.texture"
SDL_PROP_TEXTURE_CREATE_OPENGLES2_TEXTURE_UV_NUMBER = "SDL.texture.create.opengles2.texture_uv"
SDL_PROP_TEXTURE_CREATE_OPENGLES2_TEXTURE_U_NUMBER = "SDL.texture.create.opengles2.texture_u"
SDL_PROP_TEXTURE_CREATE_OPENGLES2_TEXTURE_V_NUMBER = "SDL.texture.create.opengles2.texture_v"
SDL_PROP_TEXTURE_CREATE_VULKAN_TEXTURE_NUMBER = "SDL.texture.create.vulkan.texture"

SDL_FUNC("SDL_GetTextureProperties", SDL_PropertiesID, ctypes.POINTER(SDL_Texture))

SDL_PROP_TEXTURE_COLORSPACE_NUMBER = "SDL.texture.colorspace"
SDL_PROP_TEXTURE_FORMAT_NUMBER = "SDL.texture.format"
SDL_PROP_TEXTURE_ACCESS_NUMBER = "SDL.texture.access"
SDL_PROP_TEXTURE_WIDTH_NUMBER = "SDL.texture.width"
SDL_PROP_TEXTURE_HEIGHT_NUMBER = "SDL.texture.height"
SDL_PROP_TEXTURE_SDR_WHITE_POINT_FLOAT = "SDL.texture.SDR_white_point"
SDL_PROP_TEXTURE_HDR_HEADROOM_FLOAT = "SDL.texture.HDR_headroom"
SDL_PROP_TEXTURE_D3D11_TEXTURE_POINTER = "SDL.texture.d3d11.texture"
SDL_PROP_TEXTURE_D3D11_TEXTURE_U_POINTER = "SDL.texture.d3d11.texture_u"
SDL_PROP_TEXTURE_D3D11_TEXTURE_V_POINTER = "SDL.texture.d3d11.texture_v"
SDL_PROP_TEXTURE_D3D12_TEXTURE_POINTER = "SDL.texture.d3d12.texture"
SDL_PROP_TEXTURE_D3D12_TEXTURE_U_POINTER = "SDL.texture.d3d12.texture_u"
SDL_PROP_TEXTURE_D3D12_TEXTURE_V_POINTER = "SDL.texture.d3d12.texture_v"
SDL_PROP_TEXTURE_OPENGL_TEXTURE_NUMBER = "SDL.texture.opengl.texture"
SDL_PROP_TEXTURE_OPENGL_TEXTURE_UV_NUMBER = "SDL.texture.opengl.texture_uv"
SDL_PROP_TEXTURE_OPENGL_TEXTURE_U_NUMBER = "SDL.texture.opengl.texture_u"
SDL_PROP_TEXTURE_OPENGL_TEXTURE_V_NUMBER = "SDL.texture.opengl.texture_v"
SDL_PROP_TEXTURE_OPENGL_TEXTURE_TARGET_NUMBER = "SDL.texture.opengl.target"
SDL_PROP_TEXTURE_OPENGL_TEX_W_FLOAT = "SDL.texture.opengl.tex_w"
SDL_PROP_TEXTURE_OPENGL_TEX_H_FLOAT = "SDL.texture.opengl.tex_h"
SDL_PROP_TEXTURE_OPENGLES2_TEXTURE_NUMBER = "SDL.texture.opengles2.texture"
SDL_PROP_TEXTURE_OPENGLES2_TEXTURE_UV_NUMBER = "SDL.texture.opengles2.texture_uv"
SDL_PROP_TEXTURE_OPENGLES2_TEXTURE_U_NUMBER = "SDL.texture.opengles2.texture_u"
SDL_PROP_TEXTURE_OPENGLES2_TEXTURE_V_NUMBER = "SDL.texture.opengles2.texture_v"
SDL_PROP_TEXTURE_OPENGLES2_TEXTURE_TARGET_NUMBER = "SDL.texture.opengles2.target"
SDL_PROP_TEXTURE_VULKAN_TEXTURE_NUMBER = "SDL.texture.vulkan.texture"

SDL_FUNC("SDL_GetRendererFromTexture", ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Texture))
SDL_FUNC("SDL_GetTextureSize", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float))

SDL_FUNC("SDL_SetTextureColorMod", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8)
SDL_FUNC("SDL_SetTextureColorModFloat", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.c_float, ctypes.c_float, ctypes.c_float)

SDL_FUNC("SDL_GetTextureColorMod", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8))
SDL_FUNC("SDL_GetTextureColorModFloat", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float))

SDL_FUNC("SDL_SetTextureAlphaMod", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.c_uint8)
SDL_FUNC("SDL_SetTextureAlphaModFloat", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.c_float)

SDL_FUNC("SDL_GetTextureAlphaMod", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.POINTER(ctypes.c_uint8))
SDL_FUNC("SDL_GetTextureAlphaModFloat", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.POINTER(ctypes.c_float))

SDL_FUNC("SDL_SetTextureBlendMode", ctypes.c_bool, ctypes.POINTER(SDL_Texture), SDL_BlendMode)
SDL_FUNC("SDL_GetTextureBlendMode", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_BlendMode))

SDL_FUNC("SDL_SetTextureScaleMode", ctypes.c_bool, ctypes.POINTER(SDL_Texture), SDL_ScaleMode)
SDL_FUNC("SDL_GetTextureScaleMode", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_ScaleMode))

SDL_FUNC("SDL_UpdateTexture", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_Rect), ctypes.c_void_p, ctypes.c_int)
SDL_FUNC("SDL_UpdateYUVTexture", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_Rect), ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p, ctypes.c_int)
SDL_FUNC("SDL_UpdateNVTexture", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_Rect), ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p, ctypes.c_int)

SDL_FUNC("SDL_LockTexture", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_Rect), ctypes.POINTER(ctypes.c_void_p), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_LockTextureToSurface", ctypes.c_bool, ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_Rect), ctypes.POINTER(ctypes.POINTER(SDL_Surface)))
SDL_FUNC("SDL_UnlockTexture", None, ctypes.POINTER(SDL_Texture))

SDL_FUNC("SDL_SetRenderTarget", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Texture))
SDL_FUNC("SDL_GetRenderTarget", ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_Renderer))

SDL_FUNC("SDL_SetRenderLogicalPresentation", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.c_int, ctypes.c_int, SDL_RendererLogicalPresentation, SDL_ScaleMode)
SDL_FUNC("SDL_GetRenderLogicalPresentation", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(SDL_RendererLogicalPresentation), ctypes.POINTER(SDL_ScaleMode))
SDL_FUNC("SDL_GetRenderLogicalPresentationRect", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Rect))

SDL_FUNC("SDL_RenderCoordinatesFromWindow", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.c_float, ctypes.c_float, ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float))
SDL_FUNC("SDL_RenderCoordinatesToWindow", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.c_float, ctypes.c_float, ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float))

SDL_FUNC("SDL_ConvertEventToRenderCoordinates", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Event))

SDL_FUNC("SDL_SetRenderViewport", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Rect))
SDL_FUNC("SDL_GetRenderViewport", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Rect))

SDL_FUNC("SDL_RenderViewportSet", ctypes.c_bool, ctypes.POINTER(SDL_Renderer))
SDL_FUNC("SDL_GetRenderSafeArea", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Rect))

SDL_FUNC("SDL_SetRenderClipRect", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Rect))
SDL_FUNC("SDL_GetRenderClipRect", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Rect))

SDL_FUNC("SDL_RenderClipEnabled", ctypes.c_bool, ctypes.POINTER(SDL_Renderer))

SDL_FUNC("SDL_SetRenderScale", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_GetRenderScale", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float))

SDL_FUNC("SDL_SetRenderDrawColor", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8)
SDL_FUNC("SDL_SetRenderDrawColorFloat", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float)

SDL_FUNC("SDL_GetRenderDrawColor", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8))
SDL_FUNC("SDL_GetRenderDrawColorFloat", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float))

SDL_FUNC("SDL_SetRenderColorScale", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.c_float)
SDL_FUNC("SDL_GetRenderColorScale", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(ctypes.c_float))

SDL_FUNC("SDL_SetRenderDrawBlendMode", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), SDL_BlendMode)
SDL_FUNC("SDL_GetRenderDrawBlendMode", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_BlendMode))

SDL_FUNC("SDL_RenderClear", ctypes.c_bool, ctypes.POINTER(SDL_Renderer))

SDL_FUNC("SDL_RenderPoint", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_RenderPoints", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_FPoint), ctypes.c_int)

SDL_FUNC("SDL_RenderLine", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_RenderLines", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_FPoint), ctypes.c_int)

SDL_FUNC("SDL_RenderRect", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_FRect))
SDL_FUNC("SDL_RenderRects", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_FRect), ctypes.c_int)

SDL_FUNC("SDL_RenderFillRect", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_FRect))
SDL_FUNC("SDL_RenderFillRects", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_FRect), ctypes.c_int)

SDL_FUNC("SDL_RenderTexture", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_FRect), ctypes.POINTER(SDL_FRect))
SDL_FUNC("SDL_RenderTextureRotated", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_FRect), ctypes.POINTER(SDL_FRect), ctypes.c_double, ctypes.POINTER(SDL_FPoint), SDL_FlipMode)
SDL_FUNC("SDL_RenderTextureAffine", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_FRect), ctypes.POINTER(SDL_FPoint), ctypes.POINTER(SDL_FPoint), ctypes.POINTER(SDL_FPoint))
SDL_FUNC("SDL_RenderTextureTiled", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_FRect), ctypes.c_float, ctypes.POINTER(SDL_FRect))
SDL_FUNC("SDL_RenderTexture9Grid", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_FRect), ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.POINTER(SDL_FRect))

SDL_FUNC("SDL_RenderGeometry", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Texture), ctypes.POINTER(SDL_Vertex), ctypes.c_int, ctypes.POINTER(ctypes.c_int), ctypes.c_int)
SDL_FUNC("SDL_RenderGeometryRaw", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Texture), ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.POINTER(SDL_FColor), ctypes.c_int, ctypes.POINTER(ctypes.c_float), ctypes.c_int, ctypes.c_int, ctypes.c_void_p, ctypes.c_int, ctypes.c_int)

SDL_FUNC("SDL_RenderReadPixels", ctypes.POINTER(SDL_Surface), ctypes.POINTER(SDL_Renderer), ctypes.POINTER(SDL_Rect))
SDL_FUNC("SDL_RenderPresent", ctypes.c_bool, ctypes.POINTER(SDL_Renderer))

SDL_FUNC("SDL_DestroyTexture", None, ctypes.POINTER(SDL_Texture))
SDL_FUNC("SDL_DestroyRenderer", None, ctypes.POINTER(SDL_Renderer))
SDL_FUNC("SDL_FlushRenderer", ctypes.c_bool, ctypes.POINTER(SDL_Renderer))

SDL_FUNC("SDL_GetRenderMetalLayer", ctypes.c_void_p, ctypes.POINTER(SDL_Renderer))
SDL_FUNC("SDL_GetRenderMetalCommandEncoder", ctypes.c_void_p, ctypes.POINTER(SDL_Renderer))

SDL_FUNC("SDL_AddVulkanRenderSemaphores", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.c_uint32, ctypes.c_int64, ctypes.c_int64)

SDL_FUNC("SDL_SetRenderVSync", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.c_int)

SDL_RENDERER_VSYNC_DISABLED = 0
SDL_RENDERER_VSYNC_ADAPTIVE = -1

SDL_FUNC("SDL_GetRenderVSync", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.POINTER(ctypes.c_int))

SDL_DEBUG_TEXT_FONT_CHARACTER_SIZE = 8

SDL_FUNC("SDL_RenderDebugText", ctypes.c_bool, ctypes.POINTER(SDL_Renderer), ctypes.c_float, ctypes.c_float, ctypes.c_char_p)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_DLL, SDL_DLL

SDL_SET_CURRENT_DLL(SDL_DLL)

SDL_LogCategory = ctypes.c_int

SDL_LOG_CATEGORY_APPLICATION = 0
SDL_LOG_CATEGORY_ERROR = 1
SDL_LOG_CATEGORY_ASSERT = 2
SDL_LOG_CATEGORY_SYSTEM = 3
SDL_LOG_CATEGORY_AUDIO = 4
SDL_LOG_CATEGORY_VIDEO = 5
SDL_LOG_CATEGORY_RENDER = 6
SDL_LOG_CATEGORY_INPUT = 7
SDL_LOG_CATEGORY_TEST = 8
SDL_LOG_CATEGORY_GPU = 9
SDL_LOG_CATEGORY_RESERVED1 = 10
SDL_LOG_CATEGORY_RESERVED2 = 11
SDL_LOG_CATEGORY_RESERVED3 = 12
SDL_LOG_CATEGORY_RESERVED4 = 13
SDL_LOG_CATEGORY_RESERVED5 = 14
SDL_LOG_CATEGORY_RESERVED6 = 15
SDL_LOG_CATEGORY_RESERVED7 = 16
SDL_LOG_CATEGORY_RESERVED8  = 17
SDL_LOG_CATEGORY_RESERVED9 = 18
SDL_LOG_CATEGORY_RESERVED10 = 19
SDL_LOG_CATEGORY_CUSTOM = 20

SDL_LogPriority = ctypes.c_int

SDL_LOG_PRIORITY_INVALID = 0
SDL_LOG_PRIORITY_TRACE = 1
SDL_LOG_PRIORITY_VERBOSE = 2
SDL_LOG_PRIORITY_DEBUG = 3
SDL_LOG_PRIORITY_INFO = 4
SDL_LOG_PRIORITY_WARN = 5
SDL_LOG_PRIORITY_ERROR = 6
SDL_LOG_PRIORITY_CRITICAL = 7
SDL_LOG_PRIORITY_COUNT = 8

SDL_FUNC("SDL_SetLogPriorities", None, SDL_LogPriority)
SDL_FUNC("SDL_SetLogPriority", None, ctypes.c_int, SDL_LogPriority)
SDL_FUNC("SDL_GetLogPriority", SDL_LogPriority, ctypes.c_int)
SDL_FUNC("SDL_ResetLogPriorities", None)
SDL_FUNC("SDL_SetLogPriorityPrefix", ctypes.c_bool, SDL_LogPriority, ctypes.c_char_p)

SDL_FUNC("SDL_Log", None, ctypes.c_char_p)
SDL_FUNC("SDL_LogTrace", None, ctypes.c_int, ctypes.c_char_p)
SDL_FUNC("SDL_LogVerbose", None, ctypes.c_int, ctypes.c_char_p)
SDL_FUNC("SDL_LogDebug", None, ctypes.c_int, ctypes.c_char_p)
SDL_FUNC("SDL_LogInfo", None, ctypes.c_int, ctypes.c_char_p)
SDL_FUNC("SDL_LogWarn", None, ctypes.c_int, ctypes.c_char_p)
SDL_FUNC("SDL_LogError", None, ctypes.c_int, ctypes.c_char_p)
SDL_FUNC("SDL_LogCritical", None, ctypes.c_int, ctypes.c_char_p)
SDL_FUNC("SDL_LogMessage", None, ctypes.c_int, SDL_LogPriority, ctypes.c_char_p)
SDL_FUNC("SDL_LogMessageV", None, ctypes.c_int, SDL_LogPriority, ctypes.c_char_p, ctypes.c_char_p)

SDL_LogOutputFunction = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.c_int, SDL_LogPriority, ctypes.c_char_p)

SDL_FUNC("SDL_GetDefaultLogOutputFunction", SDL_LogOutputFunction)
SDL_FUNC("SDL_GetLogOutputFunction", None, ctypes.POINTER(SDL_LogOutputFunction), ctypes.POINTER(ctypes.c_void_p))
SDL_FUNC("SDL_SetLogOutputFunction", None, SDL_LogOutputFunction, ctypes.c_void_p)


from .__init__ import ctypes, sys, \
    SDL_FUNC, SDL_SET_CURRENT_DLL, SDL_GET_DLL, SDL_TTF_DLL, SDL_DLL

from .SDL_pixels import SDL_Color, SDL_FColor
from .SDL_gpu import SDL_GPUDevice, SDL_GPUTexture
from .SDL_properties import SDL_PropertiesID
from .SDL_render import SDL_Renderer
from .SDL_surface import SDL_Surface
from .SDL_iostream import SDL_IOStream
from .SDL_version import SDL_VERSIONNUM
from .SDL_rect import SDL_Rect

SDL_SET_CURRENT_DLL(SDL_TTF_DLL)

SDL_TTF_MAJOR_VERSION = 3
SDL_TTF_MINOR_VERSION = 0
SDL_TTF_MICRO_VERSION = 0

SDL_TTF_VERSION = \
    SDL_VERSIONNUM(SDL_TTF_MAJOR_VERSION, SDL_TTF_MINOR_VERSION, SDL_TTF_MICRO_VERSION)

SDL_TTF_VERSION_ATLEAST = lambda x, y, z: \
    (SDL_TTF_MAJOR_VERSION >= x) and \
    (SDL_TTF_MAJOR_VERSION > x or SDL_TTF_MINOR_VERSION >= y) and \
    (SDL_TTF_MAJOR_VERSION > x or SDL_TTF_MINOR_VERSION > y or SDL_TTF_MICRO_VERSION >= z)

SDL_FUNC("TTF_Version", ctypes.c_int)
SDL_FUNC("TTF_GetFreeTypeVersion", None, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("TTF_GetHarfBuzzVersion", None, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))

class TTF_Font(ctypes.c_void_p):
    ...

SDL_FUNC("TTF_Init", ctypes.c_int)
SDL_FUNC("TTF_OpenFont", ctypes.POINTER(TTF_Font), ctypes.c_char_p, ctypes.c_float)
SDL_FUNC("TTF_OpenFontIO", ctypes.POINTER(TTF_Font), ctypes.POINTER(SDL_IOStream), ctypes.c_bool, ctypes.c_float)
SDL_FUNC("TTF_OpenFontWithProperties", ctypes.POINTER(TTF_Font), SDL_PropertiesID)

TTF_PROP_FONT_CREATE_FILENAME_STRING = "SDL_ttf.font.create.filename"
TTF_PROP_FONT_CREATE_IOSTREAM_POINTER = "SDL_ttf.font.create.iostream"
TTF_PROP_FONT_CREATE_IOSTREAM_OFFSET_NUMBER = "SDL_ttf.font.create.iostream.offset"
TTF_PROP_FONT_CREATE_IOSTREAM_AUTOCLOSE_BOOLEAN = "SDL_ttf.font.create.iostream.autoclose"
TTF_PROP_FONT_CREATE_SIZE_FLOAT = "SDL_ttf.font.create.size"
TTF_PROP_FONT_CREATE_FACE_NUMBER = "SDL_ttf.font.create.face"
TTF_PROP_FONT_CREATE_HORIZONTAL_DPI_NUMBER = "SDL_ttf.font.create.hdpi"
TTF_PROP_FONT_CREATE_VERTICAL_DPI_NUMBER = "SDL_ttf.font.create.vdpi"

SDL_FUNC("TTF_GetFontProperties", SDL_PropertiesID, ctypes.POINTER(TTF_Font))

TTF_PROP_FONT_OUTLINE_LINE_CAP_NUMBER = "SDL_ttf.font.outline.line_cap"
TTF_PROP_FONT_OUTLINE_LINE_JOIN_NUMBER = "SDL_ttf.font.outline.line_join"
TTF_PROP_FONT_OUTLINE_MITER_LIMIT_NUMBER = "SDL_ttf.font.outline.miter_limit"

SDL_FUNC("TTF_GetFontGeneration", ctypes.c_uint32, ctypes.POINTER(TTF_Font))

SDL_FUNC("TTF_SetFontSize", ctypes.c_bool, ctypes.POINTER(TTF_Font), ctypes.c_float)
SDL_FUNC("TTF_SetFontSizeDPI", ctypes.c_bool, ctypes.POINTER(TTF_Font), ctypes.c_float, ctypes.c_int, ctypes.c_int)

SDL_FUNC("TTF_GetFontSize", ctypes.c_float, ctypes.POINTER(TTF_Font))
SDL_FUNC("TTF_GetFontDPI", ctypes.c_bool, ctypes.POINTER(TTF_Font), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))

TTF_STYLE_NORMAL = 0x00
TTF_STYLE_BOLD = 0x01
TTF_STYLE_ITALIC = 0x02
TTF_STYLE_UNDERLINE = 0x04
TTF_STYLE_STRIKETHROUGH = 0x08

SDL_FUNC("TTF_SetFontStyle", None, ctypes.POINTER(TTF_Font), ctypes.c_int)
SDL_FUNC("TTF_GetFontStyle", ctypes.c_int, ctypes.POINTER(TTF_Font))
SDL_FUNC("TTF_SetFontOutline", ctypes.c_bool, ctypes.POINTER(TTF_Font), ctypes.c_int)
SDL_FUNC("TTF_GetFontOutline", ctypes.c_int, ctypes.POINTER(TTF_Font))

TTF_HINTING_NORMAL = 0
TTF_HINTING_LIGHT = 1
TTF_HINTING_MONO = 2
TTF_HINTING_NONE = 3
TTF_HINTING_LIGHT_SUBPIXEL = 4

SDL_FUNC("TTF_SetFontHinting", None, ctypes.POINTER(TTF_Font), ctypes.c_int)
SDL_FUNC("TTF_GetFontHinting", ctypes.c_int, ctypes.POINTER(TTF_Font))

SDL_FUNC("TTF_SetFontSDF", ctypes.c_bool, ctypes.POINTER(TTF_Font), ctypes.c_bool)
SDL_FUNC("TTF_GetFontSDF", ctypes.c_bool, ctypes.POINTER(TTF_Font))

TTF_HorizontalAlignment = ctypes.c_int

TTF_HORIZONTAL_ALIGN_INVALID = -1
TTF_HORIZONTAL_ALIGN_LEFT = 0
TTF_HORIZONTAL_ALIGN_CENTER = 1
TTF_HORIZONTAL_ALIGN_RIGHT = 2

SDL_FUNC("TTF_SetFontWrapAlignment", None, ctypes.POINTER(TTF_Font), TTF_HorizontalAlignment)
SDL_FUNC("TTF_GetFontWrapAlignment", TTF_HorizontalAlignment, ctypes.POINTER(TTF_Font))

SDL_FUNC("TTF_GetFontHeight", ctypes.c_int, ctypes.POINTER(TTF_Font))
SDL_FUNC("TTF_GetFontAscent", ctypes.c_int, ctypes.POINTER(TTF_Font))
SDL_FUNC("TTF_GetFontDescent", ctypes.c_int, ctypes.POINTER(TTF_Font))

SDL_FUNC("TTF_SetFontLineSkip", None, ctypes.POINTER(TTF_Font), ctypes.c_int)
SDL_FUNC("TTF_GetFontLineSkip", ctypes.c_int, ctypes.POINTER(TTF_Font))

SDL_FUNC("TTF_SetFontKerning", None, ctypes.POINTER(TTF_Font), ctypes.c_bool)
SDL_FUNC("TTF_GetFontKerning", ctypes.c_bool, ctypes.POINTER(TTF_Font))

SDL_FUNC("TTF_FontIsFixedWidth", ctypes.c_bool, ctypes.POINTER(TTF_Font))
SDL_FUNC("TTF_FontIsScalable", ctypes.c_bool, ctypes.POINTER(TTF_Font))

SDL_FUNC("TTF_GetFontFamilyName", ctypes.c_char_p, ctypes.POINTER(TTF_Font))
SDL_FUNC("TTF_GetFontStyleName", ctypes.c_char_p, ctypes.POINTER(TTF_Font))

TTF_Direction = ctypes.c_int

TTF_DIRECTION_LTR = 0
TTF_DIRECTION_RTL = 1
TTF_DIRECTION_TTB = 2
TTF_DIRECTION_BTT = 3

SDL_FUNC("TTF_RenderText_Solid", ctypes.POINTER(SDL_Surface), ctypes.POINTER(TTF_Font), ctypes.c_char_p, ctypes.c_size_t, SDL_Color)
SDL_FUNC("TTF_RenderText_Solid_Wrapped", ctypes.POINTER(SDL_Surface), ctypes.POINTER(TTF_Font), ctypes.c_char_p, ctypes.c_size_t, SDL_Color, ctypes.c_int)
SDL_FUNC("TTF_RenderGlyph_Solid", ctypes.POINTER(SDL_Surface), ctypes.POINTER(TTF_Font), ctypes.c_uint32, SDL_Color)

SDL_FUNC("TTF_SetFontDirection", ctypes.c_bool, ctypes.POINTER(TTF_Font), TTF_Direction)
SDL_FUNC("TTF_GetFontDirection", TTF_Direction, ctypes.POINTER(TTF_Font))
SDL_FUNC("TTF_SetFontScript", ctypes.c_bool, ctypes.POINTER(TTF_Font), ctypes.c_char_p)
SDL_FUNC("TTF_GetGlyphScript", ctypes.c_bool, ctypes.c_uint32, ctypes.c_char_p, ctypes.c_size_t)
SDL_FUNC("TTF_SetFontLanguage", ctypes.c_bool, ctypes.POINTER(TTF_Font), ctypes.c_char_p)
SDL_FUNC("TTF_FontHasGlyph", ctypes.c_bool, ctypes.POINTER(TTF_Font), ctypes.c_uint32)
SDL_FUNC("TTF_GetGlyphImage", ctypes.POINTER(SDL_Surface), ctypes.POINTER(TTF_Font), ctypes.c_uint32)
SDL_FUNC("TTF_GetGlyphImageForIndex", ctypes.POINTER(SDL_Surface), ctypes.POINTER(TTF_Font), ctypes.c_uint32)
SDL_FUNC("TTF_GetGlyphMetrics", ctypes.c_bool, ctypes.POINTER(TTF_Font), ctypes.c_uint32, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("TTF_GetGlyphKerning", ctypes.c_bool, ctypes.POINTER(TTF_Font), ctypes.c_uint32, ctypes.c_uint32, ctypes.POINTER(ctypes.c_int))
SDL_FUNC("TTF_GetStringSize", ctypes.c_bool, ctypes.POINTER(TTF_Font), ctypes.c_char_p, ctypes.c_size_t, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("TTF_GetStringSizeWrapped", ctypes.c_bool, ctypes.POINTER(TTF_Font), ctypes.c_char_p, ctypes.c_size_t, ctypes.c_int, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("TTF_MeasureString", ctypes.c_bool, ctypes.POINTER(TTF_Font), ctypes.c_char_p, ctypes.c_size_t, ctypes.c_int, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_size_t))

SDL_FUNC("TTF_RenderText_Shaded", ctypes.POINTER(SDL_Surface), ctypes.POINTER(TTF_Font), ctypes.c_char_p, ctypes.c_size_t, SDL_Color, SDL_Color)
SDL_FUNC("TTF_RenderText_Shaded_Wrapped", ctypes.POINTER(SDL_Surface), ctypes.POINTER(TTF_Font), ctypes.c_char_p, ctypes.c_size_t, SDL_Color, SDL_Color, ctypes.c_int)
SDL_FUNC("TTF_RenderGlyph_Shaded", ctypes.POINTER(SDL_Surface), ctypes.POINTER(TTF_Font), ctypes.c_uint32, SDL_Color, SDL_Color)
SDL_FUNC("TTF_RenderText_Blended", ctypes.POINTER(SDL_Surface), ctypes.POINTER(TTF_Font), ctypes.c_char_p, ctypes.c_size_t, SDL_Color)
SDL_FUNC("TTF_RenderText_Blended_Wrapped", ctypes.POINTER(SDL_Surface), ctypes.POINTER(TTF_Font), ctypes.c_char_p, ctypes.c_size_t, SDL_Color, ctypes.c_int)
SDL_FUNC("TTF_RenderGlyph_Blended", ctypes.POINTER(SDL_Surface), ctypes.POINTER(TTF_Font), ctypes.c_uint32, SDL_Color)
SDL_FUNC("TTF_RenderText_LCD", ctypes.POINTER(SDL_Surface), ctypes.POINTER(TTF_Font), ctypes.c_char_p, ctypes.c_size_t, SDL_Color, SDL_Color)
SDL_FUNC("TTF_RenderText_LCD_Wrapped", ctypes.POINTER(SDL_Surface), ctypes.POINTER(TTF_Font), ctypes.c_char_p, ctypes.c_size_t, SDL_Color, SDL_Color, ctypes.c_int)
SDL_FUNC("TTF_RenderGlyph_LCD", ctypes.POINTER(SDL_Surface), ctypes.POINTER(TTF_Font), ctypes.c_uint32, SDL_Color, SDL_Color)

class TTF_TextEngine(ctypes.c_void_p):
    ...

class TTF_TextData(ctypes.c_void_p):
    ...

class TTF_Text(ctypes.Structure):
    _fields_ = [
        ("text", ctypes.c_char_p),
        ("num_lines", ctypes.c_int),
        ("refcount", ctypes.c_int),
        ("internal", ctypes.POINTER(TTF_TextData))
    ]

SDL_FUNC("TTF_CreateSurfaceTextEngine", ctypes.POINTER(TTF_TextEngine))
SDL_FUNC("TTF_DrawSurfaceText", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.c_int, ctypes.c_int, ctypes.POINTER(SDL_Surface))
SDL_FUNC("TTF_DestroySurfaceTextEngine", None, ctypes.POINTER(TTF_TextEngine))
SDL_FUNC("TTF_CreateRendererTextEngine", ctypes.POINTER(TTF_TextEngine), ctypes.POINTER(SDL_Renderer))
SDL_FUNC("TTF_DrawRendererText", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.c_float, ctypes.c_float)
SDL_FUNC("TTF_DestroyRendererTextEngine", None, ctypes.POINTER(TTF_TextEngine))

SDL_FUNC("TTF_CreateGPUTextEngine", ctypes.POINTER(TTF_TextEngine), ctypes.POINTER(SDL_GPUDevice))

class TTF_GPUAtlasDrawSequence(ctypes.c_void_p):
    ...

class TTF_GPUAtlasDrawSequence(ctypes.Structure):
    _fields_ = [
        ("atlas_texture", ctypes.POINTER(SDL_GPUTexture)),
        ("xy", ctypes.POINTER(ctypes.c_float)),
        ("xy_stride", ctypes.c_int),
        ("uv", ctypes.POINTER(ctypes.c_float)),
        ("uv_stride", ctypes.c_int),
        ("num_vertices", ctypes.c_int),
        ("indices", ctypes.POINTER(ctypes.c_int)),
        ("num_indices", ctypes.c_int),
        ("next", ctypes.POINTER(TTF_GPUAtlasDrawSequence))
    ]

SDL_FUNC("TTF_GetGPUTextDrawData", ctypes.POINTER(TTF_GPUAtlasDrawSequence), ctypes.POINTER(TTF_Text))
SDL_FUNC("TTF_DestroyGPUTextEngine", None, ctypes.POINTER(TTF_TextEngine))

SDL_FUNC("TTF_CreateText", ctypes.POINTER(TTF_Text), ctypes.POINTER(TTF_TextEngine), ctypes.POINTER(TTF_Font), ctypes.c_char_p, ctypes.c_size_t)

if "win32" in sys.platform:
    SDL_FUNC("TTF_GetTextProperties", SDL_PropertiesID, ctypes.POINTER(TTF_Text))

SDL_FUNC("TTF_SetTextEngine", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.POINTER(TTF_TextEngine))
SDL_FUNC("TTF_GetTextEngine", ctypes.POINTER(TTF_TextEngine), ctypes.POINTER(TTF_Text))
SDL_FUNC("TTF_SetTextFont", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.POINTER(TTF_Font))
SDL_FUNC("TTF_GetTextFont", ctypes.POINTER(TTF_Font), ctypes.POINTER(TTF_Text))
SDL_FUNC("TTF_SetTextColor", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8)
SDL_FUNC("TTF_SetTextColorFloat", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.c_float, ctypes.c_float, ctypes.c_float, ctypes.c_float)
SDL_FUNC("TTF_GetTextColor", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8))
SDL_FUNC("TTF_GetTextColorFloat", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float))
SDL_FUNC("TTF_SetTextPosition", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.c_int, ctypes.c_int)
SDL_FUNC("TTF_GetTextPosition", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("TTF_SetTextWrapWidth", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.c_int)
SDL_FUNC("TTF_GetTextWrapWidth", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("TTF_SetTextWrapWhitespaceVisible", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.c_bool)
SDL_FUNC("TTF_TextWrapWhitespaceVisible", ctypes.c_bool, ctypes.POINTER(TTF_Text))
SDL_FUNC("TTF_SetTextString", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.c_char_p, ctypes.c_size_t)
SDL_FUNC("TTF_InsertTextString", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.c_int, ctypes.c_char_p, ctypes.c_size_t)
SDL_FUNC("TTF_AppendTextString", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.c_char_p, ctypes.c_size_t)
SDL_FUNC("TTF_DeleteTextString", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.c_int, ctypes.c_int)
SDL_FUNC("TTF_GetTextSize", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))

TTF_SubStringFlags = ctypes.c_uint32

TTF_SUBSTRING_TEXT_STARTD = 0x00000001  
TTF_SUBSTRING_LINE_STARTD = 0x00000002  
TTF_SUBSTRING_LINE_ENDD = 0x00000004  
TTF_SUBSTRING_TEXT_END = 0x00000008  

class TTF_SubString(ctypes.Structure):
    _fields_ = [
        ("flags", TTF_SubStringFlags),
        ("offset", ctypes.c_int),
        ("length", ctypes.c_int),
        ("line_index", ctypes.c_int),
        ("cluster_index", ctypes.c_int),
        ("rect", SDL_Rect)
    ]

SDL_FUNC("TTF_GetTextSubString", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.c_int, ctypes.POINTER(TTF_SubString))
SDL_FUNC("TTF_GetTextSubStringForLine", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.c_int, ctypes.POINTER(TTF_SubString))
SDL_FUNC("TTF_GetTextSubStringsForRange", ctypes.POINTER(ctypes.POINTER(TTF_SubString)), ctypes.POINTER(TTF_Text), ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_int))
SDL_FUNC("TTF_GetTextSubStringForPoint", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.c_int, ctypes.c_int, ctypes.POINTER(TTF_SubString))
SDL_FUNC("TTF_GetPreviousTextSubString", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.POINTER(TTF_SubString), ctypes.POINTER(TTF_SubString))
SDL_FUNC("TTF_GetNextTextSubString", ctypes.c_bool, ctypes.POINTER(TTF_Text), ctypes.POINTER(TTF_SubString), ctypes.POINTER(TTF_SubString))
SDL_FUNC("TTF_UpdateText", None, ctypes.POINTER(TTF_Text))
SDL_FUNC("TTF_DestroyText", None, ctypes.POINTER(TTF_Text))

SDL_FUNC("TTF_CloseFont", None, ctypes.POINTER(TTF_Font))
SDL_FUNC("TTF_Quit", None)
SDL_FUNC("TTF_WasInit", ctypes.c_int)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_DEREFERENCE, SDL_SET_CURRENT_DLL, SDL_DLL

from .SDL_stdinc import SDL_FLT_EPSILON

SDL_SET_CURRENT_DLL(SDL_DLL)

class SDL_Point(ctypes.Structure):
    _fields_ = [
        ("x", ctypes.c_int),
        ("y", ctypes.c_int)
    ]
    
class SDL_FPoint(ctypes.Structure):
    _fields_ = [
        ("x", ctypes.c_float),
        ("y", ctypes.c_float)
    ]

class SDL_Rect(ctypes.Structure):
    _fields_ = [
        ("x", ctypes.c_int),
        ("y", ctypes.c_int),
        ("w", ctypes.c_int),
        ("h", ctypes.c_int)
    ]

class SDL_FRect(ctypes.Structure):
    _fields_ = [
        ("x", ctypes.c_float),
        ("y", ctypes.c_float),
        ("w", ctypes.c_float),
        ("h", ctypes.c_float)
    ]

class LP_SDL_Point(ctypes.POINTER(SDL_Point)):
    ...

class LP_SDL_FPoint(ctypes.POINTER(SDL_FPoint)):
    ...

class LP_SDL_Rect(ctypes.POINTER(SDL_Rect)):
    ...

class LP_SDL_FRect(ctypes.POINTER(SDL_FRect)):
    ...

def SDL_RectToFRect(rect: LP_SDL_Rect, frect: LP_SDL_FRect) -> None:
    rect, frect = SDL_DEREFERENCE(rect), SDL_DEREFERENCE(frect)
    frect.x, frect.y, frect.w, frect.h = \
        float(rect.x), float(rect.y), float(rect.w), float(rect.h)
    
def SDL_PointInRect(p: LP_SDL_Point, r: LP_SDL_Rect) -> bool:
    p, r = SDL_DEREFERENCE(p), SDL_DEREFERENCE(r)
    return p.x >= r.x and p.x < r.x + r.w and p.y >= r.y and p.y < r.y + r.h

def SDL_RectEmpty(r: LP_SDL_Rect) -> bool:
    r = SDL_DEREFERENCE(r)
    return r.w <= 0 or r.h <= 0

def SDL_RectEquals(a: LP_SDL_Rect, b: LP_SDL_Rect) -> bool:
    a, b = SDL_DEREFERENCE(a), SDL_DEREFERENCE(b)
    return a.x == b.x and a.y == b.y and a.w == b.w and a.h == b.h

SDL_FUNC("SDL_HasRectIntersection", ctypes.c_bool, ctypes.POINTER(SDL_Rect), ctypes.POINTER(SDL_Rect))
SDL_FUNC("SDL_GetRectIntersection", ctypes.c_bool, ctypes.POINTER(SDL_Rect), ctypes.POINTER(SDL_Rect), ctypes.POINTER(SDL_Rect))
SDL_FUNC("SDL_GetRectUnion", ctypes.c_bool, ctypes.POINTER(SDL_Rect), ctypes.POINTER(SDL_Rect), ctypes.POINTER(SDL_Rect))
SDL_FUNC("SDL_GetRectEnclosingPoints", ctypes.c_bool, ctypes.POINTER(SDL_Point), ctypes.c_int, ctypes.POINTER(SDL_Rect), ctypes.POINTER(SDL_Rect))
SDL_FUNC("SDL_GetRectAndLineIntersection", ctypes.c_bool, ctypes.POINTER(SDL_Rect), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))

def SDL_PointInRectFloat(p: LP_SDL_FPoint, r: LP_SDL_FRect) -> bool:
    p, r = SDL_DEREFERENCE(p), SDL_DEREFERENCE(r)
    return p.x >= r.x and p.x < r.x + r.w and p.y >= r.y and p.y < r.y + r.h

def SDL_RectEmptyFloat(r: LP_SDL_FRect) -> bool:
    r = SDL_DEREFERENCE(r)
    return r.w <= 0 or r.h <= 0

def SDL_RectsEqualEpsilon(a: LP_SDL_FRect, b: LP_SDL_FRect, epsilon: float) -> bool:
    a, b = SDL_DEREFERENCE(a), SDL_DEREFERENCE(b)
    return abs(a.x - b.x) < epsilon and abs(a.y - b.y) < epsilon and abs(a.w - b.w) < epsilon and abs(a.h - b.h) < epsilon

def SDL_RectsEqualFloat(a: LP_SDL_FRect, b: LP_SDL_FRect) -> bool:
    a, b = SDL_DEREFERENCE(a), SDL_DEREFERENCE(b)
    return SDL_RectsEqualEpsilon(a, b, SDL_FLT_EPSILON)

SDL_FUNC("SDL_HasRectIntersectionFloat", ctypes.c_bool, ctypes.POINTER(SDL_FRect), ctypes.POINTER(SDL_FRect))
SDL_FUNC("SDL_GetRectIntersectionFloat", ctypes.c_bool, ctypes.POINTER(SDL_FRect), ctypes.POINTER(SDL_FRect), ctypes.POINTER(SDL_FRect))
SDL_FUNC("SDL_GetRectUnionFloat", ctypes.c_bool, ctypes.POINTER(SDL_FRect), ctypes.POINTER(SDL_FRect), ctypes.POINTER(SDL_FRect))
SDL_FUNC("SDL_GetRectEnclosingPointsFloat", ctypes.c_bool, ctypes.POINTER(SDL_FPoint), ctypes.c_int, ctypes.POINTER(SDL_FRect), ctypes.POINTER(SDL_FRect))
SDL_FUNC("SDL_GetRectAndLineIntersectionFloat", ctypes.c_bool, ctypes.POINTER(SDL_FRect), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float), ctypes.POINTER(ctypes.c_float))


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_DLL, SDL_DLL

SDL_SET_CURRENT_DLL(SDL_DLL)

SDL_PropertiesID = ctypes.c_uint32
SDL_PropertyType = ctypes.c_int

SDL_PROPERTY_TYPE_INVALID = 0
SDL_PROPERTY_TYPE_POINTER = 1
SDL_PROPERTY_TYPE_STRING = 2
SDL_PROPERTY_TYPE_NUMBER = 3
SDL_PROPERTY_TYPE_FLOAT = 4
SDL_PROPERTY_TYPE_BOOLEAN = 5

SDL_FUNC("SDL_GetGlobalProperties", SDL_PropertiesID)
SDL_FUNC("SDL_CreateProperties", SDL_PropertiesID)
SDL_FUNC("SDL_CopyProperties", ctypes.c_bool, SDL_PropertiesID, SDL_PropertiesID)
SDL_FUNC("SDL_LockProperties", ctypes.c_bool, SDL_PropertiesID)
SDL_FUNC("SDL_UnlockProperties", None, SDL_PropertiesID)

SDL_CleanupPropertyCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.c_void_p)

SDL_FUNC("SDL_SetPointerPropertyWithCleanup", ctypes.c_bool, SDL_PropertiesID, ctypes.c_char_p, ctypes.c_void_p, SDL_CleanupPropertyCallback, ctypes.c_void_p)
SDL_FUNC("SDL_SetPointerProperty", ctypes.c_bool, SDL_PropertiesID, ctypes.c_char_p, ctypes.c_void_p)
SDL_FUNC("SDL_SetStringProperty", ctypes.c_bool, SDL_PropertiesID, ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_SetNumberProperty", ctypes.c_bool, SDL_PropertiesID, ctypes.c_char_p, ctypes.c_int64)
SDL_FUNC("SDL_SetFloatProperty", ctypes.c_bool, SDL_PropertiesID, ctypes.c_char_p, ctypes.c_float)
SDL_FUNC("SDL_SetBooleanProperty", ctypes.c_bool, SDL_PropertiesID, ctypes.c_char_p, ctypes.c_bool)
SDL_FUNC("SDL_HasProperty", ctypes.c_bool, SDL_PropertiesID, ctypes.c_char_p)
SDL_FUNC("SDL_GetPropertyType", SDL_PropertyType, SDL_PropertiesID, ctypes.c_char_p)
SDL_FUNC("SDL_GetPointerProperty", ctypes.c_void_p, SDL_PropertiesID, ctypes.c_char_p, ctypes.c_void_p)
SDL_FUNC("SDL_GetStringProperty", ctypes.c_char_p, SDL_PropertiesID, ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_GetNumberProperty", ctypes.c_int64, SDL_PropertiesID, ctypes.c_char_p, ctypes.c_int64)
SDL_FUNC("SDL_GetFloatProperty", ctypes.c_float, SDL_PropertiesID, ctypes.c_char_p, ctypes.c_float)
SDL_FUNC("SDL_GetBooleanProperty", ctypes.c_bool, SDL_PropertiesID, ctypes.c_char_p, ctypes.c_bool)
SDL_FUNC("SDL_ClearProperty", ctypes.c_bool, SDL_PropertiesID, ctypes.c_char_p)

SDL_EnumeratePropertiesCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, SDL_PropertiesID, ctypes.c_char_p)

SDL_FUNC("SDL_EnumerateProperties", ctypes.c_bool, SDL_PropertiesID, SDL_EnumeratePropertiesCallback, ctypes.c_void_p)
SDL_FUNC("SDL_DestroyProperties", None, SDL_PropertiesID)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_DLL, SDL_MIXER_DLL

from .SDL_iostream import SDL_IOStream
from .SDL_audio import SDL_AudioDeviceID, SDL_AudioSpec, SDL_AudioFormat, SDL_AUDIO_S16
from .SDL_version import SDL_VERSIONNUM

SDL_SET_CURRENT_DLL(SDL_MIXER_DLL)

SDL_MIXER_MAJOR_VERSION = 3
SDL_MIXER_MINOR_VERSION = 0
SDL_MIXER_MICRO_VERSION = 0

SDL_MIXER_VERSION = \
    SDL_VERSIONNUM(SDL_MIXER_MAJOR_VERSION, SDL_MIXER_MINOR_VERSION, SDL_MIXER_MICRO_VERSION)

SDL_MIXER_VERSION_ATLEAST = lambda x, y, z: \
    (SDL_MIXER_MAJOR_VERSION >= x) and \
    (SDL_MIXER_MAJOR_VERSION > x or SDL_MIXER_MINOR_VERSION >= y) and \
    (SDL_MIXER_MAJOR_VERSION > x or SDL_MIXER_MINOR_VERSION > y or SDL_MIXER_MICRO_VERSION >= z)

SDL_FUNC("Mix_Version", ctypes.c_int)

MIX_InitFlags = ctypes.c_uint32

MIX_INIT_FLAC = 0x00000001
MIX_INIT_MOD = 0x00000002
MIX_INIT_MP3 = 0x00000008
MIX_INIT_OGG = 0x00000010
MIX_INIT_MID = 0x00000020
MIX_INIT_OPUS = 0x00000040
MIX_INIT_WAVPACK = 0x00000080

SDL_FUNC("Mix_Init", MIX_InitFlags, MIX_InitFlags)
SDL_FUNC("Mix_Quit", None)

MIX_CHANNELS = 8

MIX_DEFAULT_FREQUENCY = 44100
MIX_DEFAULT_FORMAT = SDL_AUDIO_S16
MIX_DEFAULT_CHANNELS = 2
MIX_MAX_VOLUME = 128 

class Mix_Chunk(ctypes.Structure):
    _fields_ = [
        ("allocated", ctypes.c_int),
        ("abuf", ctypes.POINTER(ctypes.c_uint8)),
        ("alen", ctypes.c_uint32),
        ("volume", ctypes.c_uint8)
    ]

Mix_Fading = ctypes.c_int

MIX_NO_FADING = 0
MIX_FADING_OUT = 1
MIX_FADING_IN = 2

Mix_MusicType = ctypes.c_int

MUS_NONE = 0
MUS_WAV = 1
MUS_MOD = 2
MUS_MID = 3
MUS_OGG = 4
MUS_MP3 = 5
MUS_MP3_MAD_UNUSED = 6
MUS_FLAC = 7
MUS_MODPLUG_UNUSED = 8
MUS_OPUS = 9
MUS_WAVPACK = 10
MUS_GME = 11

class Mix_Music(ctypes.c_void_p):
    ...

SDL_FUNC("Mix_OpenAudio", ctypes.c_bool, SDL_AudioDeviceID, ctypes.POINTER(SDL_AudioSpec))
SDL_FUNC("Mix_PauseAudio", None, ctypes.c_int)
SDL_FUNC("Mix_QuerySpec", ctypes.c_bool, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(SDL_AudioFormat), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("Mix_AllocateChannels", ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_LoadWAV_IO", ctypes.POINTER(Mix_Chunk), ctypes.POINTER(SDL_IOStream), ctypes.c_bool)
SDL_FUNC("Mix_LoadWAV", ctypes.POINTER(Mix_Chunk), ctypes.c_char_p)
SDL_FUNC("Mix_LoadMUS", ctypes.POINTER(Mix_Music), ctypes.c_char_p)
SDL_FUNC("Mix_LoadMUS_IO", ctypes.POINTER(Mix_Music), ctypes.POINTER(SDL_IOStream), ctypes.c_bool)
SDL_FUNC("Mix_LoadMUSType_IO", ctypes.POINTER(Mix_Music), ctypes.POINTER(SDL_IOStream), Mix_MusicType, ctypes.c_bool)
SDL_FUNC("Mix_QuickLoad_WAV", ctypes.POINTER(Mix_Chunk), ctypes.POINTER(ctypes.c_uint8))
SDL_FUNC("Mix_QuickLoad_RAW", ctypes.POINTER(Mix_Chunk), ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32)
SDL_FUNC("Mix_FreeChunk", None, ctypes.POINTER(Mix_Chunk))
SDL_FUNC("Mix_FreeMusic", None, ctypes.POINTER(Mix_Music))
SDL_FUNC("Mix_GetNumChunkDecoders", ctypes.c_int)
SDL_FUNC("Mix_GetChunkDecoder", ctypes.c_char_p, ctypes.c_int)
SDL_FUNC("Mix_HasChunkDecoder", ctypes.c_bool, ctypes.c_char_p)
SDL_FUNC("Mix_GetNumMusicDecoders", ctypes.c_int)
SDL_FUNC("Mix_GetMusicDecoder", ctypes.c_char_p, ctypes.c_int)
SDL_FUNC("Mix_HasMusicDecoder", ctypes.c_bool, ctypes.c_char_p)
SDL_FUNC("Mix_GetMusicType", Mix_MusicType, ctypes.POINTER(Mix_Music))
SDL_FUNC("Mix_GetMusicTitle", ctypes.c_char_p, ctypes.POINTER(Mix_Music))
SDL_FUNC("Mix_GetMusicTitleTag", ctypes.c_char_p, ctypes.POINTER(Mix_Music))
SDL_FUNC("Mix_GetMusicArtistTag", ctypes.c_char_p, ctypes.POINTER(Mix_Music))
SDL_FUNC("Mix_GetMusicAlbumTag", ctypes.c_char_p, ctypes.POINTER(Mix_Music))
SDL_FUNC("Mix_GetMusicCopyrightTag", ctypes.c_char_p, ctypes.POINTER(Mix_Music))

Mix_MixCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(ctypes.c_uint8), ctypes.c_int)

SDL_FUNC("Mix_SetPostMix", None, Mix_MixCallback, ctypes.c_void_p)
SDL_FUNC("Mix_HookMusic", None, Mix_MixCallback, ctypes.c_void_p)

Mix_MusicFinishedCallback = ctypes.CFUNCTYPE(None)

SDL_FUNC("Mix_HookMusicFinished", None, Mix_MusicFinishedCallback)
SDL_FUNC("Mix_GetMusicHookData", ctypes.c_void_p)

Mix_ChannelFinishedCallback = ctypes.CFUNCTYPE(None, ctypes.c_int)

SDL_FUNC("Mix_ChannelFinished", None, Mix_ChannelFinishedCallback)

MIX_CHANNEL_POST = -2

Mix_EffectFunc_t = ctypes.CFUNCTYPE(None, ctypes.c_int, ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p)
Mix_EffectDone_t = ctypes.CFUNCTYPE(None, ctypes.c_int, ctypes.c_void_p)

SDL_FUNC("Mix_RegisterEffect", ctypes.c_bool, ctypes.c_int, Mix_EffectFunc_t, Mix_EffectDone_t, ctypes.c_void_p)
SDL_FUNC("Mix_UnregisterEffect", ctypes.c_bool, ctypes.c_int, Mix_EffectFunc_t)
SDL_FUNC("Mix_UnregisterAllEffects", ctypes.c_bool, ctypes.c_int)

MIX_EFFECTSMAXSPEED = "MIX_EFFECTSMAXSPEED"

SDL_FUNC("Mix_SetPanning", ctypes.c_bool, ctypes.c_int, ctypes.c_uint8, ctypes.c_uint8)
SDL_FUNC("Mix_SetPosition", ctypes.c_bool, ctypes.c_int, ctypes.c_int16, ctypes.c_uint8)
SDL_FUNC("Mix_SetDistance", ctypes.c_bool, ctypes.c_int, ctypes.c_uint8)
SDL_FUNC("Mix_SetReverseStereo", ctypes.c_bool, ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_ReserveChannels", ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_GroupChannel", ctypes.c_bool, ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_GroupChannels", ctypes.c_bool, ctypes.c_int, ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_GroupAvailable", ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_GroupCount", ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_GroupOldest", ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_GroupNewer", ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_PlayChannel", ctypes.c_int, ctypes.c_int, ctypes.POINTER(Mix_Chunk), ctypes.c_int)
SDL_FUNC("Mix_PlayChannelTimed", ctypes.c_int, ctypes.c_int, ctypes.POINTER(Mix_Chunk), ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_PlayMusic", ctypes.c_bool, ctypes.POINTER(Mix_Music), ctypes.c_int)    
SDL_FUNC("Mix_FadeInMusic", ctypes.c_bool, ctypes.POINTER(Mix_Music), ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_FadeInMusicPos", ctypes.c_bool, ctypes.POINTER(Mix_Music), ctypes.c_int, ctypes.c_int, ctypes.c_double)
SDL_FUNC("Mix_FadeInChannel", ctypes.c_int, ctypes.c_int, ctypes.POINTER(Mix_Chunk), ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_FadeInChannelTimed", ctypes.c_int, ctypes.c_int, ctypes.POINTER(Mix_Chunk), ctypes.c_int, ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_Volume", ctypes.c_int, ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_VolumeChunk", ctypes.c_int, ctypes.POINTER(Mix_Chunk), ctypes.c_int)
SDL_FUNC("Mix_VolumeMusic", ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_GetMusicVolume", ctypes.c_int, ctypes.POINTER(Mix_Music))
SDL_FUNC("Mix_MasterVolume", ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_HaltChannel", None, ctypes.c_int)
SDL_FUNC("Mix_HaltGroup", None, ctypes.c_int)
SDL_FUNC("Mix_HaltMusic", None)
SDL_FUNC("Mix_ExpireChannel", ctypes.c_int, ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_FadeOutChannel", ctypes.c_int, ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_FadeOutGroup", ctypes.c_int, ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_FadeOutMusic", ctypes.c_bool, ctypes.c_int)
SDL_FUNC("Mix_FadingMusic", Mix_Fading)
SDL_FUNC("Mix_FadingChannel", Mix_Fading, ctypes.c_int)
SDL_FUNC("Mix_Pause", None, ctypes.c_int)
SDL_FUNC("Mix_PauseGroup", None, ctypes.c_int)
SDL_FUNC("Mix_Resume", None, ctypes.c_int)
SDL_FUNC("Mix_ResumeGroup", None, ctypes.c_int)
SDL_FUNC("Mix_Paused", ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_PauseMusic", None)
SDL_FUNC("Mix_ResumeMusic", None)
SDL_FUNC("Mix_RewindMusic", None)
SDL_FUNC("Mix_PausedMusic", ctypes.c_bool)
SDL_FUNC("Mix_ModMusicJumpToOrder", ctypes.c_bool, ctypes.c_int)
SDL_FUNC("Mix_StartTrack", ctypes.c_bool, ctypes.POINTER(Mix_Music), ctypes.c_int)
SDL_FUNC("Mix_GetNumTracks", ctypes.c_int, ctypes.POINTER(Mix_Music))
SDL_FUNC("Mix_SetMusicPosition", ctypes.c_bool, ctypes.c_double)
SDL_FUNC("Mix_GetMusicPosition", ctypes.c_double, ctypes.POINTER(Mix_Music))
SDL_FUNC("Mix_MusicDuration", ctypes.c_double, ctypes.POINTER(Mix_Music))
SDL_FUNC("Mix_GetMusicLoopStartTime", ctypes.c_double, ctypes.POINTER(Mix_Music))
SDL_FUNC("Mix_GetMusicLoopEndTime", ctypes.c_double, ctypes.POINTER(Mix_Music))
SDL_FUNC("Mix_GetMusicLoopLengthTime", ctypes.c_double, ctypes.POINTER(Mix_Music))
SDL_FUNC("Mix_Playing", ctypes.c_int, ctypes.c_int)
SDL_FUNC("Mix_PlayingMusic", ctypes.c_bool)
SDL_FUNC("Mix_SetSoundFonts", ctypes.c_bool, ctypes.c_char_p)
SDL_FUNC("Mix_GetSoundFonts", ctypes.c_char_p)

Mix_EachSoundFontCallback = ctypes.CFUNCTYPE(ctypes.c_bool, ctypes.c_char_p, ctypes.c_void_p)

SDL_FUNC("Mix_EachSoundFont", ctypes.c_bool, Mix_EachSoundFontCallback, ctypes.c_void_p)
SDL_FUNC("Mix_SetTimidityCfg", ctypes.c_bool, ctypes.c_char_p)
SDL_FUNC("Mix_GetTimidityCfg", ctypes.c_char_p)
SDL_FUNC("Mix_GetChunk", ctypes.POINTER(Mix_Chunk), ctypes.c_int)
SDL_FUNC("Mix_CloseAudio", None)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_DLL, SDL_DLL

from .SDL_endian import SDL_BYTEORDER, SDL_LIL_ENDIAN
from .SDL_properties import SDL_PropertiesID
from .SDL_iostream import SDL_IOStream

SDL_SET_CURRENT_DLL(SDL_DLL)

SDL_AUDIO_MASK_BITSIZE = 0xFF
SDL_AUDIO_MASK_FLOAT = 1 << 8
SDL_AUDIO_MASK_BIG_ENDIAN = 1 << 12
SDL_AUDIO_MASK_SIGNED = 1 << 15

SDL_DEFINE_AUDIO_FORMAT = lambda signed, bigendian, _float, size: \
    (signed << 15) | (bigendian << 12) | (_float << 8) | (size & SDL_AUDIO_MASK_BITSIZE)

SDL_AudioFormat = ctypes.c_int

SDL_AUDIO_UNKNOWN = 0x0000
SDL_AUDIO_U8 = 0x0008
SDL_AUDIO_S8 = 0x8008
SDL_AUDIO_S16LE = 0x8010
SDL_AUDIO_S16BE = 0x9010
SDL_AUDIO_S32LE = 0x8020
SDL_AUDIO_S32BE = 0x9020
SDL_AUDIO_F32LE = 0x8120
SDL_AUDIO_F32BE = 0x9120

if SDL_BYTEORDER == SDL_LIL_ENDIAN:
    SDL_AUDIO_S16 = SDL_AUDIO_S16LE
    SDL_AUDIO_S32 = SDL_AUDIO_S32LE
    SDL_AUDIO_F32 = SDL_AUDIO_F32LE

else:
    SDL_AUDIO_S16 = SDL_AUDIO_S16BE
    SDL_AUDIO_S32 = SDL_AUDIO_S32BE
    SDL_AUDIO_F32 = SDL_AUDIO_F32BE

SDL_AUDIO_BITSIZE = lambda x: x & SDL_AUDIO_MASK_BITSIZE
SDL_AUDIO_BYTESIZE = lambda x: SDL_AUDIO_BITSIZE(x) / 8

SDL_AUDIO_ISFLOAT = lambda x: x & SDL_AUDIO_MASK_FLOAT
SDL_AUDIO_ISBIGENDIAN = lambda x: x & SDL_AUDIO_MASK_BIG_ENDIAN
SDL_AUDIO_ISLITTLEENDIAN = lambda x: not SDL_AUDIO_ISBIGENDIAN(x)
SDL_AUDIO_ISSIGNED = lambda x: x & SDL_AUDIO_MASK_SIGNED
SDL_AUDIO_ISINT = lambda x: not SDL_AUDIO_ISFLOAT(x)
SDL_AUDIO_ISUNSIGNED = lambda x: not SDL_AUDIO_ISSIGNED(x)

SDL_AudioDeviceID = ctypes.c_uint32

SDL_AUDIO_DEVICE_DEFAULT_PLAYBACK = 0xFFFFFFFF
SDL_AUDIO_DEVICE_DEFAULT_RECORDING = 0xFFFFFFFE

class SDL_AudioSpec(ctypes.Structure):
    _fields_ = [
        ("format", SDL_AudioFormat),
        ("channels", ctypes.c_int),
        ("freq", ctypes.c_int)
    ]

SDL_AUDIO_FRAMESIZE = lambda x: SDL_AUDIO_BYTESIZE(x.format) * x.channels

class SDL_AudioStream(ctypes.c_void_p):
    ...

SDL_FUNC("SDL_GetNumAudioDrivers", ctypes.c_int)
SDL_FUNC("SDL_GetAudioDriver", ctypes.c_char_p, ctypes.c_int)
SDL_FUNC("SDL_GetCurrentAudioDriver", ctypes.c_char_p)

SDL_FUNC("SDL_GetAudioPlaybackDevices", ctypes.POINTER(SDL_AudioDeviceID), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_GetAudioRecordingDevices", ctypes.POINTER(SDL_AudioDeviceID), ctypes.POINTER(ctypes.c_int))

SDL_FUNC("SDL_GetAudioDeviceName", ctypes.c_char_p, SDL_AudioDeviceID)
SDL_FUNC("SDL_GetAudioDeviceFormat", ctypes.c_bool, SDL_AudioDeviceID, ctypes.POINTER(SDL_AudioSpec), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_GetAudioDeviceChannelMap", ctypes.POINTER(ctypes.c_int), SDL_AudioDeviceID, ctypes.POINTER(ctypes.c_int))

SDL_FUNC("SDL_OpenAudioDevice", SDL_AudioDeviceID, SDL_AudioDeviceID, ctypes.POINTER(SDL_AudioSpec))
SDL_FUNC("SDL_IsAudioDevicePhysical", ctypes.c_bool, SDL_AudioDeviceID)
SDL_FUNC("SDL_IsAudioDevicePlayback", ctypes.c_bool, SDL_AudioDeviceID)
SDL_FUNC("SDL_PauseAudioDevice", ctypes.c_bool, SDL_AudioDeviceID)
SDL_FUNC("SDL_ResumeAudioDevice", ctypes.c_bool, SDL_AudioDeviceID)
SDL_FUNC("SDL_AudioDevicePaused", ctypes.c_bool, SDL_AudioDeviceID)
SDL_FUNC("SDL_GetAudioDeviceGain", ctypes.c_float, SDL_AudioDeviceID)
SDL_FUNC("SDL_SetAudioDeviceGain", ctypes.c_bool, SDL_AudioDeviceID, ctypes.c_float)
SDL_FUNC("SDL_CloseAudioDevice", None, SDL_AudioDeviceID)

SDL_FUNC("SDL_BindAudioStreams", ctypes.c_bool, SDL_AudioDeviceID, ctypes.POINTER(ctypes.POINTER(SDL_AudioStream)), ctypes.c_int)
SDL_FUNC("SDL_BindAudioStream", ctypes.c_bool, SDL_AudioDeviceID, ctypes.POINTER(SDL_AudioStream))
SDL_FUNC("SDL_UnbindAudioStreams", None, ctypes.POINTER(ctypes.POINTER(SDL_AudioStream)), ctypes.c_int)
SDL_FUNC("SDL_UnbindAudioStream", None, ctypes.POINTER(SDL_AudioStream))
SDL_FUNC("SDL_GetAudioStreamDevice", SDL_AudioDeviceID, ctypes.POINTER(SDL_AudioStream))
SDL_FUNC("SDL_CreateAudioStream", ctypes.POINTER(SDL_AudioStream), ctypes.POINTER(SDL_AudioSpec), ctypes.POINTER(SDL_AudioSpec))
SDL_FUNC("SDL_GetAudioStreamProperties", SDL_PropertiesID, ctypes.POINTER(SDL_AudioStream))
SDL_FUNC("SDL_GetAudioStreamFormat", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream), ctypes.POINTER(SDL_AudioSpec), ctypes.POINTER(SDL_AudioSpec))
SDL_FUNC("SDL_SetAudioStreamFormat", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream), ctypes.POINTER(SDL_AudioSpec), ctypes.POINTER(SDL_AudioSpec))
SDL_FUNC("SDL_GetAudioStreamFrequencyRatio", ctypes.c_float, ctypes.POINTER(SDL_AudioStream))
SDL_FUNC("SDL_SetAudioStreamFrequencyRatio", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream), ctypes.c_float)
SDL_FUNC("SDL_GetAudioStreamGain", ctypes.c_float, ctypes.POINTER(SDL_AudioStream))
SDL_FUNC("SDL_SetAudioStreamGain", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream), ctypes.c_float)
SDL_FUNC("SDL_GetAudioStreamInputChannelMap", ctypes.POINTER(ctypes.c_int), ctypes.POINTER(SDL_AudioStream), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_GetAudioStreamOutputChannelMap", ctypes.POINTER(ctypes.c_int), ctypes.POINTER(SDL_AudioStream), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_SetAudioStreamInputChannelMap", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream), ctypes.POINTER(ctypes.c_int), ctypes.c_int)
SDL_FUNC("SDL_SetAudioStreamOutputChannelMap", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream), ctypes.POINTER(ctypes.c_int), ctypes.c_int)
SDL_FUNC("SDL_PutAudioStreamData", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream), ctypes.c_void_p, ctypes.c_int)
SDL_FUNC("SDL_GetAudioStreamData", ctypes.c_int, ctypes.POINTER(SDL_AudioStream), ctypes.c_void_p, ctypes.c_int)
SDL_FUNC("SDL_GetAudioStreamAvailable", ctypes.c_int, ctypes.POINTER(SDL_AudioStream))
SDL_FUNC("SDL_GetAudioStreamQueued", ctypes.c_int, ctypes.POINTER(SDL_AudioStream))
SDL_FUNC("SDL_FlushAudioStream", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream))
SDL_FUNC("SDL_ClearAudioStream", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream))
SDL_FUNC("SDL_PauseAudioStreamDevice", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream))
SDL_FUNC("SDL_ResumeAudioStreamDevice", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream))
SDL_FUNC("SDL_LockAudioStream", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream))
SDL_FUNC("SDL_UnlockAudioStream", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream))

SDL_AudioStreamCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(SDL_AudioStream), ctypes.c_int, ctypes.c_int)

SDL_FUNC("SDL_SetAudioStreamGetCallback", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream), SDL_AudioStreamCallback, ctypes.c_void_p)
SDL_FUNC("SDL_SetAudioStreamPutCallback", ctypes.c_bool, ctypes.POINTER(SDL_AudioStream), SDL_AudioStreamCallback, ctypes.c_void_p)
SDL_FUNC("SDL_DestroyAudioStream", None, ctypes.POINTER(SDL_AudioStream))
SDL_FUNC("SDL_OpenAudioDeviceStream", ctypes.POINTER(SDL_AudioStream), SDL_AudioDeviceID, ctypes.POINTER(SDL_AudioSpec), SDL_AudioStreamCallback, ctypes.c_void_p)

SDL_AudioPostmixCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(SDL_AudioSpec), ctypes.POINTER(ctypes.c_float), ctypes.c_int)

SDL_FUNC("SDL_SetAudioPostmixCallback", ctypes.c_bool, SDL_AudioDeviceID, SDL_AudioPostmixCallback, ctypes.c_void_p)

SDL_FUNC("SDL_LoadWAV_IO", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_bool, ctypes.POINTER(SDL_AudioSpec), ctypes.POINTER(ctypes.POINTER(ctypes.c_uint8)), ctypes.POINTER(ctypes.c_uint32))
SDL_FUNC("SDL_LoadWAV", ctypes.c_bool, ctypes.c_char_p, ctypes.POINTER(SDL_AudioSpec), ctypes.POINTER(ctypes.POINTER(ctypes.c_uint8)), ctypes.POINTER(ctypes.c_uint32))
SDL_FUNC("SDL_MixAudio", ctypes.c_bool, ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8), SDL_AudioFormat, ctypes.c_uint32, ctypes.c_float)
SDL_FUNC("SDL_ConvertAudioSamples", ctypes.c_bool, ctypes.POINTER(SDL_AudioSpec), ctypes.POINTER(ctypes.c_uint8), ctypes.c_int, ctypes.POINTER(SDL_AudioSpec), ctypes.POINTER(ctypes.POINTER(ctypes.c_uint8)), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_GetAudioFormatName", ctypes.c_char_p, SDL_AudioFormat)
SDL_FUNC("SDL_GetSilenceValueForFormat", ctypes.c_int, SDL_AudioFormat)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_DLL, SDL_DLL

SDL_SET_CURRENT_DLL(SDL_DLL)

class SDL_hid_device(ctypes.c_void_p):
    ...

SDL_hid_bus_type = ctypes.c_int

SDL_HID_API_BUS_UNKNOWN = 0x00
SDL_HID_API_BUS_USB = 0x01
SDL_HID_API_BUS_BLUETOOTH = 0x02
SDL_HID_API_BUS_I2C = 0x03
SDL_HID_API_BUS_SPI = 0x04

SDL_hid_device_info = ctypes.c_void_p

class SDL_hid_device_info(ctypes.Structure):
    _fields_ = [
        ("path", ctypes.c_char_p),
        ("vendor_id", ctypes.c_ushort),
        ("product_id", ctypes.c_ushort),
        ("serial_number", ctypes.c_wchar_p),
        ("release_number", ctypes.c_ushort),
        ("manufacturer_string", ctypes.c_wchar_p),
        ("product_string", ctypes.c_wchar_p),
        ("usage_page", ctypes.c_ushort),
        ("usage", ctypes.c_ushort),
        ("interface_number", ctypes.c_int),
        ("interface_class", ctypes.c_int),
        ("interface_subclass", ctypes.c_int),
        ("interface_protocol", ctypes.c_int),
        ("bus_type", SDL_hid_bus_type),
        ("next", ctypes.POINTER(SDL_hid_device_info))
    ]

SDL_FUNC("SDL_hid_init", ctypes.c_int)
SDL_FUNC("SDL_hid_exit", ctypes.c_int)
SDL_FUNC("SDL_hid_device_change_count", ctypes.c_uint32)
SDL_FUNC("SDL_hid_enumerate", ctypes.POINTER(SDL_hid_device_info), ctypes.c_ushort, ctypes.c_ushort)
SDL_FUNC("SDL_hid_free_enumeration", None, ctypes.POINTER(SDL_hid_device_info))
SDL_FUNC("SDL_hid_open", ctypes.POINTER(SDL_hid_device), ctypes.c_ushort, ctypes.c_ushort, ctypes.c_wchar_p)
SDL_FUNC("SDL_hid_open_path", ctypes.POINTER(SDL_hid_device), ctypes.c_char_p)
SDL_FUNC("SDL_hid_write", ctypes.c_int, ctypes.POINTER(SDL_hid_device), ctypes.POINTER(ctypes.c_ubyte), ctypes.c_size_t)
SDL_FUNC("SDL_hid_read_timeout", ctypes.c_int, ctypes.POINTER(SDL_hid_device), ctypes.POINTER(ctypes.c_ubyte), ctypes.c_size_t, ctypes.c_int)
SDL_FUNC("SDL_hid_read", ctypes.c_int, ctypes.POINTER(SDL_hid_device), ctypes.POINTER(ctypes.c_ubyte), ctypes.c_size_t)
SDL_FUNC("SDL_hid_set_nonblocking", ctypes.c_int, ctypes.POINTER(SDL_hid_device), ctypes.c_int)
SDL_FUNC("SDL_hid_send_feature_report", ctypes.c_int, ctypes.POINTER(SDL_hid_device), ctypes.POINTER(ctypes.c_ubyte), ctypes.c_size_t)
SDL_FUNC("SDL_hid_get_feature_report", ctypes.c_int, ctypes.POINTER(SDL_hid_device), ctypes.POINTER(ctypes.c_ubyte), ctypes.c_size_t)
SDL_FUNC("SDL_hid_get_input_report", ctypes.c_int, ctypes.POINTER(SDL_hid_device), ctypes.POINTER(ctypes.c_ubyte), ctypes.c_size_t)
SDL_FUNC("SDL_hid_close", ctypes.c_int, ctypes.POINTER(SDL_hid_device))
SDL_FUNC("SDL_hid_get_manufacturer_string", ctypes.c_int, ctypes.POINTER(SDL_hid_device), ctypes.c_wchar_p, ctypes.c_size_t)
SDL_FUNC("SDL_hid_get_product_string", ctypes.c_int, ctypes.POINTER(SDL_hid_device), ctypes.c_wchar_p, ctypes.c_size_t)
SDL_FUNC("SDL_hid_get_serial_number_string", ctypes.c_int, ctypes.POINTER(SDL_hid_device), ctypes.c_wchar_p, ctypes.c_size_t)
SDL_FUNC("SDL_hid_get_indexed_string", ctypes.c_int, ctypes.POINTER(SDL_hid_device), ctypes.c_int, ctypes.c_wchar_p, ctypes.c_size_t)
SDL_FUNC("SDL_hid_get_device_info", ctypes.POINTER(SDL_hid_device_info), ctypes.POINTER(SDL_hid_device))
SDL_FUNC("SDL_hid_get_report_descriptor", ctypes.c_int, ctypes.POINTER(SDL_hid_device), ctypes.POINTER(ctypes.c_ubyte), ctypes.c_size_t)
SDL_FUNC("SDL_hid_ble_scan", None, ctypes.c_bool)


"""A pure Python wrapper for SDL3."""

__version__ = "0.8.0b0"

import sys, os, requests, ctypes, platform, asyncio, aiohttp, re, inspect, array

SDL_DLL, SDL_IMAGE_DLL, SDL_MIXER_DLL, SDL_NET_DLL, SDL_RTF_DLL, SDL_TTF_DLL = \
    "SDL3", "SDL3_image", "SDL3_mixer", "SDL3_net", "SDL3_rtf", "SDL3_ttf"

SDL_DLL_VAR_MAP = {}

for i in locals().copy():
    if i.startswith("SDL_") and i.endswith("_DLL"):
        SDL_DLL_VAR_MAP[i] = locals()[i]

SDL_DLL_VAR_MAP_INV = {value: key for key, value in SDL_DLL_VAR_MAP.items()}

for key, value in {"PYSDL3_DOC_GENERATOR": "1", "PYSDL3_VERSION_CHECK": "1"}.items():
    if key not in os.environ: os.environ[key] = value

__doc_file__ = os.path.join(os.path.dirname(__file__), "__doc__.py")
__doc_generator__ = int(os.environ.get("PYSDL3_DOC_GENERATOR")) > 0

__initialized__ = __name__.split(".")[0] in sys.modules if "." in __name__ else False
__module__ = sys.modules[__name__.split(".")[0] if "." in __name__ else __name__]

def SDL_SET_TEXT_ATTR(color):
    if sys.platform in ["win32"]:
        console_handle = ctypes.windll.kernel32.GetStdHandle(-11)
        ctypes.windll.kernel32.SetConsoleTextAttribute(console_handle, color)

    else:
        if color == 7:
            print("\u001b[39m", end = "", flush = True)

        elif color == 13:
            print("\u001b[35m", end = "", flush = True)

        elif color == 12:
            print("\u001b[31m", end = "", flush = True)

        else:
            ...

if not __initialized__:
    if int(os.environ.get("PYSDL3_VERSION_CHECK")) > 0:
        try:
            version = requests.get(f"https://pypi.org/pypi/PySDL3/json", timeout = 0.5).json()["info"]["version"]

            if __version__ != version:
                SDL_SET_TEXT_ATTR(13)
                print(f"you are using an older version of pysdl3 (current: {__version__}, lastest: {version}).", flush = True)
                SDL_SET_TEXT_ATTR(7)

        except:
            ...

    functions, instances, dllMap, dll = {}, {}, {}, None
    binaryPath = os.path.join(os.path.dirname(__file__), "bin", f"{platform.system().lower()}-{platform.machine().lower().replace('x86_64', 'amd64').replace('aarch64', 'arm64')}")

    for key, value in SDL_DLL_VAR_MAP.items():
        functions[value], dllMap[value] = {}, (ctypes.WinDLL if "win32" in sys.platform else ctypes.CDLL) \
            (os.path.join(binaryPath, ("{}.dll" if "win32" in sys.platform else "lib{}.so").format(value)))

def SDL_DEREFERENCE(value):
    if isinstance(value, ctypes._Pointer):
        return value.contents
    
    elif hasattr(value, "_obj"):
        return value._obj
    
    else:
        return value

def SDL_FUNC_CACHE(func):
    cache = {}

    def wrapper(*args):
        if args in cache:
            return cache[args]
        
        else:
            cache[args] = func(*args)
            return cache[args]

    return wrapper

@SDL_FUNC_CACHE
def SDL_GET_DLL_NAME(dll):
    return {v: k for k, v in __module__.dllMap.items()}[dll]

def SDL_GET_DLL(name):
    return __module__.dllMap[name]

def SDL_SET_CURRENT_DLL(name):
    __module__.dll = SDL_GET_DLL(name)

def SDL_GET_CURRENT_DLL():
    return __module__.dll

def SDL_GET_FUNCTION_DLL(name):
    return __module__.functions[name].__dll__

def SDL_FUNC(name, retType, *args):
    func = getattr(dll := SDL_GET_CURRENT_DLL(), name)
    func.__dll__, func.restype, func.argtypes = dll, retType, args
    if not __doc_generator__: setattr(__module__, name, func)
    __module__.functions[SDL_GET_DLL_NAME(dll)][name] = func

async def SDL_GET_DESCRIPTIONS(urls):
    session, tasks = aiohttp.ClientSession(), []

    for url in urls:
        print(f"sending a request to \"{url}\".\n", end = "", flush = True)
        tasks.append(asyncio.create_task(session.get(url, ssl = False)))
    
    responses = await asyncio.gather(*tasks)
    print(f"response gathering completed ({len(responses)} response(s)).\n", end = "", flush = True)
    descriptions = []

    for response in responses:
        if response.status != 200:
            print(f"failed to get description of \"{response.url}\", skipping (status: {response.status}).\n", end = "", flush = True)
            descriptions.append(None)

        else:
            content = (await response.content.read()).decode()

            for a, b in zip(list(re.finditer(r"<p>", content)), list(re.finditer(r"</p>", content))):
                if content[a.end()] == "<": continue
                description = re.sub(r"<a href=\"([a-zA-Z0-9_]+)\">([a-zA-Z0-9_]+)</a>", rf"[\2]({'/'.join(str(response.url).split('/')[:-1])}/\1)", content[a.end():b.start()])
                descriptions.append(description.replace("<em>", "").replace("</em>", "").replace("<code>", "`").replace("</code>", "`").replace("<strong>", "").replace("</strong>", ""))
                break

        response.close()

    await session.close()
    return descriptions

def SDL_GENERATE_DOCS():
    __index, descriptions = -1, \
        asyncio.run(SDL_GET_DESCRIPTIONS([f"https://wiki.libsdl.org/{module}/{name}" for module in __module__.functions for name in __module__.functions[module]]))

    for module in __module__.functions:
        for name in __module__.functions[module]:
            __module__.functions[module][name].__doc__ = descriptions[__index := __index + 1]

    result = "\"\"\"This file is auto-generated.\"\"\"\n\n"
    result += "from .SDL import *\n\n"
    result += f"from .__init__ import ctypes, SDL_GET_DLL, \\\n"
    result += f"    {', '.join(list(SDL_DLL_VAR_MAP.keys()))}\n\n"
    types, definitions = set(), ""

    def SDL_GET_NAME(i):
        if i is None: return "None"
        if "CFunctionType" in i.__name__:
            return "ctypes._Pointer"

        if "LP_" in i.__name__ or not i.__name__.startswith("c_"):
            if "LP_" in i.__name__:
                types.add(i.__name__)

            return i.__name__
        
        else:
            return f"ctypes.{i.__name__}"

    for index, module in enumerate(__module__.functions):
        if len(__module__.functions[module]) == 0: continue
        definitions += f"# {module}.dll implementation.\n\n"

        for _index, name in enumerate(__module__.functions[module]):
            retType, args = \
                __module__.functions[module][name].restype, __module__.functions[module][name].argtypes

            definitions += f"def {name}({', '.join([f'_{index}: {SDL_GET_NAME(i)}' for index, i in enumerate(args)])}) -> {SDL_GET_NAME(retType)}:\n"
            definitions += f"    \"\"\"\n"

            if __module__.functions[module][name].__doc__ is not None and "no such page" not in __module__.functions[module][name].__doc__.lower():
                definitions += f"    {__module__.functions[module][name].__doc__}\n"

            definitions += f"    https://wiki.libsdl.org/{module}/{name}\n"
            definitions += f"    \"\"\"\n"
            definitions += f"    return SDL_GET_DLL({SDL_DLL_VAR_MAP_INV[module]}).{name}({', '.join([f'_{index}' for index, i in enumerate(args)])})"

            if _index != len(__module__.functions[module]) - 1:
                definitions += "\n\n"

        if index != len(__module__.functions) - 1 and len(list(__module__.functions.values())[index + 1]) != 0:
            definitions += "\n\n"

    for i in types:
        count, name = i.count("LP_"), i.replace("LP_", "")
        result += f"class {i}({'ctypes.POINTER(' * count}{'ctypes.' if name.startswith('c_') else ''}{name}{')' * count}):\n"
        result += f"    ...\n\n"

    return result + definitions

from .SDL import *

if not __initialized__:
    if __doc_generator__:
        if not os.path.exists(__doc_file__):
            with open(__doc_file__, "w") as file:
                file.write(SDL_GENERATE_DOCS())

        from .__doc__ import *

    else:
        if os.path.exists(__doc_file__):
            os.remove(__doc_file__)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_DLL, SDL_DLL

from .SDL_video import SDL_Window
from .SDL_properties import SDL_PropertiesID

SDL_SET_CURRENT_DLL(SDL_DLL)

class SDL_DialogFileFilter(ctypes.Structure):
    _fields_ = [
        ("name", ctypes.c_char_p),
        ("pattern", ctypes.c_char_p)
    ]

SDL_DialogFileCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(ctypes.c_char_p), ctypes.c_int)

SDL_FUNC("SDL_ShowOpenFileDialog", None, SDL_DialogFileCallback, ctypes.c_void_p, ctypes.POINTER(SDL_Window), ctypes.POINTER(SDL_DialogFileFilter), ctypes.c_int, ctypes.c_char_p, ctypes.c_bool)
SDL_FUNC("SDL_ShowSaveFileDialog", None, SDL_DialogFileCallback, ctypes.c_void_p, ctypes.POINTER(SDL_Window), ctypes.POINTER(SDL_DialogFileFilter), ctypes.c_int, ctypes.c_char_p)
SDL_FUNC("SDL_ShowOpenFolderDialog", None, SDL_DialogFileCallback, ctypes.c_void_p, ctypes.POINTER(SDL_Window), ctypes.c_char_p, ctypes.c_bool)

SDL_FileDialogType = ctypes.c_int

SDL_FILEDIALOG_OPENFILE = 0
SDL_FILEDIALOG_SAVEFILE = 1
SDL_FILEDIALOG_OPENFOLDER = 2

SDL_PROP_FILE_DIALOG_FILTERS_POINTER = "SDL.filedialog.filters"
SDL_PROP_FILE_DIALOG_NFILTERS_NUMBER = "SDL.filedialog.nfilters"
SDL_PROP_FILE_DIALOG_WINDOW_POINTER = "SDL.filedialog.window"
SDL_PROP_FILE_DIALOG_LOCATION_STRING = "SDL.filedialog.location"
SDL_PROP_FILE_DIALOG_MANY_BOOLEAN = "SDL.filedialog.many"
SDL_PROP_FILE_DIALOG_TITLE_STRING = "SDL.filedialog.title"
SDL_PROP_FILE_DIALOG_ACCEPT_STRING = "SDL.filedialog.accept"
SDL_PROP_FILE_DIALOG_CANCEL_STRING = "SDL.filedialog.cancel"

SDL_FUNC("SDL_ShowFileDialogWithProperties", None, SDL_FileDialogType, SDL_DialogFileCallback, ctypes.c_void_p, SDL_PropertiesID)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_DLL, SDL_GET_DLL, SDL_DLL

SDL_SET_CURRENT_DLL(SDL_DLL)

SDL_arraysize = lambda array: ctypes.sizeof(array) // ctypes.sizeof(array[0])

SDL_FALSE = False
SDL_TRUE = True

SDL_bool = ctypes.c_bool

Sint8 = ctypes.c_int8
SDL_MAX_SINT8 = 0x7F
SDL_MIN_SINT8 = ~0X7F

Uint8 = ctypes.c_uint8
SDL_MAX_UINT8 = 0xFF
SDL_MIN_UINT8 = 0X00

Sint16 = ctypes.c_int16
SDL_MAX_SINT16 = 0x7FFF
SDL_MIN_SINT16 = ~0X7FFF

Uint16 = ctypes.c_uint16
SDL_MAX_UINT16 = 0xFFFF
SDL_MIN_UINT16 = 0X0000

Sint32 = ctypes.c_int32
SDL_MAX_SINT32 = 0x7FFFFFFF
SDL_MIN_SINT32 = ~0X7FFFFFFF

Uint32 = ctypes.c_uint32
SDL_MAX_UINT32 = 0xFFFFFFFF
SDL_MIN_UINT32 = 0X00000000

Sint64 = ctypes.c_int64
SDL_MAX_SINT64 = 0x7FFFFFFFFFFFFFFF
SDL_MIN_SINT64 = ~0X7FFFFFFFFFFFFFFF

Uint64 = ctypes.c_uint64
SDL_MAX_UINT64 = 0xFFFFFFFFFFFFFFFF
SDL_MIN_UINT64 = 0X0000000000000000

SDL_Time = Sint64
SDL_MAX_TIME = SDL_MAX_SINT64
SDL_MIN_TIME = SDL_MIN_SINT64

SDL_FLT_EPSILON = 1.1920928955078125e-07

SDL_FUNC("SDL_malloc", ctypes.c_void_p, ctypes.c_size_t)
SDL_FUNC("SDL_calloc", ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t)
SDL_FUNC("SDL_realloc", ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t)
SDL_FUNC("SDL_free", None, ctypes.c_void_p)

SDL_malloc_func = ctypes.CFUNCTYPE(ctypes.c_void_p, ctypes.c_size_t)
SDL_calloc_func = ctypes.CFUNCTYPE(ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t)
SDL_realloc_func = ctypes.CFUNCTYPE(ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t)
SDL_free_func = ctypes.CFUNCTYPE(None, ctypes.c_void_p)

SDL_FUNC("SDL_GetOriginalMemoryFunctions", None, ctypes.POINTER(SDL_malloc_func), ctypes.POINTER(SDL_calloc_func), ctypes.POINTER(SDL_realloc_func), ctypes.POINTER(SDL_free_func))
SDL_FUNC("SDL_GetMemoryFunctions", None, ctypes.POINTER(SDL_malloc_func), ctypes.POINTER(SDL_calloc_func), ctypes.POINTER(SDL_realloc_func), ctypes.POINTER(SDL_free_func))
SDL_FUNC("SDL_SetMemoryFunctions", ctypes.c_bool, SDL_malloc_func, SDL_calloc_func, SDL_realloc_func, SDL_free_func)

SDL_FUNC("SDL_aligned_alloc", ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t)
SDL_FUNC("SDL_aligned_free", None, ctypes.c_void_p)

SDL_FUNC("SDL_GetNumAllocations", ctypes.c_int)

class SDL_Environment(ctypes.c_void_p):
    ...

SDL_FUNC("SDL_GetEnvironment", ctypes.POINTER(SDL_Environment))
SDL_FUNC("SDL_CreateEnvironment", ctypes.POINTER(SDL_Environment), ctypes.c_bool)
SDL_FUNC("SDL_GetEnvironmentVariable", ctypes.c_char_p, ctypes.POINTER(SDL_Environment), ctypes.c_char_p)
SDL_FUNC("SDL_GetEnvironmentVariables", ctypes.POINTER(ctypes.c_char_p), ctypes.POINTER(SDL_Environment))
SDL_FUNC("SDL_SetEnvironmentVariable", ctypes.c_bool, ctypes.POINTER(SDL_Environment), ctypes.c_char_p, ctypes.c_char_p, ctypes.c_bool)
SDL_FUNC("SDL_UnsetEnvironmentVariable", ctypes.c_bool, ctypes.POINTER(SDL_Environment), ctypes.c_char_p)
SDL_FUNC("SDL_DestroyEnvironment", None, ctypes.POINTER(SDL_Environment))

SDL_FUNC("SDL_getenv", ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_getenv_unsafe", ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_setenv_unsafe", ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_unsetenv_unsafe", ctypes.c_char_p, ctypes.c_char_p)

SDL_CompareCallback = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.c_void_p, ctypes.c_void_p)

SDL_FUNC("SDL_qsort", None, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t, SDL_CompareCallback)
SDL_FUNC("SDL_bsearch", ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t, SDL_CompareCallback)

SDL_CompareCallback_r = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p)

SDL_FUNC("SDL_qsort_r", None, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t, SDL_CompareCallback_r, ctypes.c_void_p)
SDL_FUNC("SDL_bsearch_r", ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_size_t, SDL_CompareCallback_r, ctypes.c_void_p)

SDL_FUNC("SDL_abs", ctypes.c_int, ctypes.c_int)

SDL_min = lambda x, y: x if x < y else y
SDL_max = lambda x, y: x if x > y else y
SDL_clamp = lambda x, a, b: a if x < a else (b if x > b else x)

SDL_FUNC("SDL_isalpha", ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_isalnum", ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_isblank", ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_iscntrl", ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_isdigit", ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_isxdigit", ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_ispunct", ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_isspace", ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_isupper", ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_islower", ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_isprint", ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_isgraph", ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_toupper", ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_tolower", ctypes.c_int, ctypes.c_int)

SDL_FUNC("SDL_crc16", ctypes.c_uint16, ctypes.c_uint16, ctypes.c_void_p, ctypes.c_size_t)
SDL_FUNC("SDL_crc32", ctypes.c_uint32, ctypes.c_uint32, ctypes.c_void_p, ctypes.c_size_t)
SDL_FUNC("SDL_murmur3_32", ctypes.c_uint32, ctypes.c_void_p, ctypes.c_size_t, ctypes.c_uint32)

SDL_FUNC("SDL_memcpy", ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t)
SDL_FUNC("SDL_memmove", ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t)
SDL_FUNC("SDL_memset", ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_size_t)
SDL_FUNC("SDL_memset4", ctypes.c_void_p, ctypes.c_void_p, ctypes.c_uint32, ctypes.c_size_t)

SDL_zero = lambda x: SDL_GET_DLL(SDL_DLL).SDL_memset(ctypes.byref(x), 0, ctypes.sizeof(x))
SDL_zerop = lambda x: SDL_GET_DLL(SDL_DLL).SDL_memset(x, 0, ctypes.sizeof(x.contents))
SDL_zeroa = lambda x: SDL_GET_DLL(SDL_DLL).SDL_memset(x, 0, ctypes.sizeof(x))

SDL_FUNC("SDL_memcmp", ctypes.c_int, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t)

SDL_FUNC("SDL_wcslen", ctypes.c_size_t, ctypes.c_wchar_p)
SDL_FUNC("SDL_wcsnlen", ctypes.c_size_t, ctypes.c_wchar_p, ctypes.c_size_t)
SDL_FUNC("SDL_wcslcpy", ctypes.c_size_t, ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_size_t)
SDL_FUNC("SDL_wcslcat", ctypes.c_size_t, ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_size_t)
SDL_FUNC("SDL_wcsdup", ctypes.c_wchar_p, ctypes.c_wchar_p)
SDL_FUNC("SDL_wcsstr", ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_wchar_p)
SDL_FUNC("SDL_wcsnstr", ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_size_t)
SDL_FUNC("SDL_wcscmp", ctypes.c_int, ctypes.c_wchar_p, ctypes.c_wchar_p)
SDL_FUNC("SDL_wcsncmp", ctypes.c_int, ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_size_t)
SDL_FUNC("SDL_wcscasecmp", ctypes.c_int, ctypes.c_wchar_p, ctypes.c_wchar_p)
SDL_FUNC("SDL_wcsncasecmp", ctypes.c_int, ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_size_t)
SDL_FUNC("SDL_wcstol", ctypes.c_long, ctypes.c_wchar_p, ctypes.POINTER(ctypes.c_wchar_p), ctypes.c_int)

SDL_FUNC("SDL_strlen", ctypes.c_size_t, ctypes.c_char_p)
SDL_FUNC("SDL_strnlen", ctypes.c_size_t, ctypes.c_char_p, ctypes.c_size_t)
SDL_FUNC("SDL_strlcpy", ctypes.c_size_t, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t)
SDL_FUNC("SDL_utf8strlcpy", ctypes.c_size_t, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t)
SDL_FUNC("SDL_strlcat", ctypes.c_size_t, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t)
SDL_FUNC("SDL_strdup", ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_strndup", ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t)
SDL_FUNC("SDL_strrev", ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_strupr", ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_strlwr", ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_strchr", ctypes.c_char_p, ctypes.c_char_p, ctypes.c_int)
SDL_FUNC("SDL_strrchr", ctypes.c_char_p, ctypes.c_char_p, ctypes.c_int)
SDL_FUNC("SDL_strstr", ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_strnstr", ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t)
SDL_FUNC("SDL_strcasestr", ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_strtok_r", ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.POINTER(ctypes.c_char_p))
SDL_FUNC("SDL_utf8strlen", ctypes.c_size_t, ctypes.c_char_p)
SDL_FUNC("SDL_utf8strnlen", ctypes.c_size_t, ctypes.c_char_p, ctypes.c_size_t)

SDL_FUNC("SDL_itoa", ctypes.c_char_p, ctypes.c_int, ctypes.c_char_p, ctypes.c_int)
SDL_FUNC("SDL_uitoa", ctypes.c_char_p, ctypes.c_uint, ctypes.c_char_p, ctypes.c_int)
SDL_FUNC("SDL_ltoa", ctypes.c_char_p, ctypes.c_long, ctypes.c_char_p, ctypes.c_int)
SDL_FUNC("SDL_ultoa", ctypes.c_char_p, ctypes.c_ulong, ctypes.c_char_p, ctypes.c_int)
SDL_FUNC("SDL_lltoa", ctypes.c_char_p, ctypes.c_int64, ctypes.c_char_p, ctypes.c_int)
SDL_FUNC("SDL_ulltoa", ctypes.c_char_p, ctypes.c_uint64, ctypes.c_char_p, ctypes.c_int)

SDL_FUNC("SDL_atoi", ctypes.c_int, ctypes.c_char_p)
SDL_FUNC("SDL_atof", ctypes.c_double, ctypes.c_char_p)
SDL_FUNC("SDL_strtol", ctypes.c_long, ctypes.c_char_p, ctypes.POINTER(ctypes.c_char_p), ctypes.c_int)
SDL_FUNC("SDL_strtoul", ctypes.c_ulong, ctypes.c_char_p, ctypes.POINTER(ctypes.c_char_p), ctypes.c_int)
SDL_FUNC("SDL_strtoll", ctypes.c_int64, ctypes.c_char_p, ctypes.POINTER(ctypes.c_char_p), ctypes.c_int)
SDL_FUNC("SDL_strtoull", ctypes.c_uint64, ctypes.c_char_p, ctypes.POINTER(ctypes.c_char_p), ctypes.c_int)
SDL_FUNC("SDL_strtod", ctypes.c_double, ctypes.c_char_p, ctypes.POINTER(ctypes.c_char_p))

SDL_FUNC("SDL_strcmp", ctypes.c_int, ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_strncmp", ctypes.c_int, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t)
SDL_FUNC("SDL_strcasecmp", ctypes.c_int, ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_strncasecmp", ctypes.c_int, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t)

SDL_INVALID_UNICODE_CODEPOINT = 0xFFFD

SDL_FUNC("SDL_StepUTF8", ctypes.c_uint32, ctypes.POINTER(ctypes.c_char_p), ctypes.POINTER(ctypes.c_size_t))
SDL_FUNC("SDL_StepBackUTF8", ctypes.c_uint32, ctypes.c_char_p, ctypes.POINTER(ctypes.c_char_p))
SDL_FUNC("SDL_UCS4ToUTF8", ctypes.c_char_p, ctypes.c_uint32, ctypes.c_char_p)

SDL_FUNC("SDL_sscanf", ctypes.c_int, ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_vsscanf", ctypes.c_int, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_void_p)
SDL_FUNC("SDL_snprintf", ctypes.c_int, ctypes.c_char_p, ctypes.c_size_t, ctypes.c_char_p)
SDL_FUNC("SDL_swprintf", ctypes.c_int, ctypes.c_wchar_p, ctypes.c_size_t, ctypes.c_wchar_p)
SDL_FUNC("SDL_vsnprintf", ctypes.c_int, ctypes.c_char_p, ctypes.c_size_t, ctypes.c_char_p, ctypes.c_void_p)
SDL_FUNC("SDL_vswprintf", ctypes.c_int, ctypes.c_wchar_p, ctypes.c_size_t, ctypes.c_wchar_p, ctypes.c_void_p)
SDL_FUNC("SDL_asprintf", ctypes.c_int, ctypes.POINTER(ctypes.c_char_p), ctypes.c_char_p)
SDL_FUNC("SDL_vasprintf", ctypes.c_int, ctypes.POINTER(ctypes.c_char_p), ctypes.c_char_p, ctypes.c_void_p)

SDL_FUNC("SDL_srand", None, ctypes.c_uint64)
SDL_FUNC("SDL_rand", ctypes.c_int32, ctypes.c_int64)
SDL_FUNC("SDL_randf", ctypes.c_float)
SDL_FUNC("SDL_rand_bits", ctypes.c_uint32)
SDL_FUNC("SDL_rand_r", ctypes.c_int32, ctypes.POINTER(ctypes.c_uint64), ctypes.c_int32)
SDL_FUNC("SDL_randf_r", ctypes.c_float, ctypes.POINTER(ctypes.c_uint64))
SDL_FUNC("SDL_rand_bits_r", ctypes.c_uint32, ctypes.POINTER(ctypes.c_uint64))

SDL_PI_D = SDL_PI_F = 3.141592653589793238462643383279502884

SDL_FUNC("SDL_acos", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_acosf", ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_asin", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_asinf", ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_atan", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_atanf", ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_atan2", ctypes.c_double, ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_atan2f", ctypes.c_float, ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_ceil", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_ceilf", ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_copysign", ctypes.c_double, ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_copysignf", ctypes.c_float, ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_cos", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_cosf", ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_exp", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_expf", ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_fabs", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_fabsf", ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_floor", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_floorf", ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_trunc", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_truncf", ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_fmod", ctypes.c_double, ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_fmodf", ctypes.c_float, ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_isinf", ctypes.c_int, ctypes.c_double)
SDL_FUNC("SDL_isinff", ctypes.c_int, ctypes.c_float)
SDL_FUNC("SDL_isnan", ctypes.c_int, ctypes.c_double)
SDL_FUNC("SDL_isnanf", ctypes.c_int, ctypes.c_float)
SDL_FUNC("SDL_log", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_logf", ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_log10", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_log10f", ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_modf", ctypes.c_double, ctypes.c_double, ctypes.POINTER(ctypes.c_double))
SDL_FUNC("SDL_modff", ctypes.c_float, ctypes.c_float, ctypes.POINTER(ctypes.c_float))
SDL_FUNC("SDL_pow", ctypes.c_double, ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_powf", ctypes.c_float, ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_round", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_roundf", ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_lround", ctypes.c_long, ctypes.c_double)
SDL_FUNC("SDL_lroundf", ctypes.c_long, ctypes.c_float)
SDL_FUNC("SDL_scalbn", ctypes.c_double, ctypes.c_double, ctypes.c_int)
SDL_FUNC("SDL_scalbnf", ctypes.c_float, ctypes.c_float, ctypes.c_int)
SDL_FUNC("SDL_sin", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_sinf", ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_sqrt", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_sqrtf", ctypes.c_float, ctypes.c_float)
SDL_FUNC("SDL_tan", ctypes.c_double, ctypes.c_double)
SDL_FUNC("SDL_tanf", ctypes.c_float, ctypes.c_float)

SDL_ICONV_ERROR = -1
SDL_ICONV_E2BIG = -2
SDL_ICONV_EILSEQ = -3
SDL_ICONV_EINVAL = -4

class SDL_iconv_t(ctypes.c_void_p):
    ...

SDL_FUNC("SDL_iconv_open", SDL_iconv_t, ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_iconv_close", ctypes.c_int, SDL_iconv_t)
SDL_FUNC("SDL_iconv", ctypes.c_size_t, SDL_iconv_t, ctypes.POINTER(ctypes.c_char_p), ctypes.POINTER(ctypes.c_size_t), ctypes.POINTER(ctypes.c_char_p), ctypes.POINTER(ctypes.c_size_t))
SDL_FUNC("SDL_iconv_string", ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_size_t)

SDL_iconv_utf8_locale = lambda s: \
    SDL_GET_DLL(SDL_DLL).SDL_iconv_string("".encode(), "UTF-8".encode(), s.encode(), SDL_GET_DLL(SDL_DLL).SDL_strlen(s.encode()) + 1)

SDL_iconv_utf8_ucs2 = lambda s: \
    SDL_GET_DLL(SDL_DLL).SDL_iconv_string("UCS-2".encode(), "UTF-8".encode(), s.encode(), SDL_GET_DLL(SDL_DLL).SDL_strlen(s.encode()) + 1)

SDL_iconv_utf8_ucs4 = lambda s: \
    SDL_GET_DLL(SDL_DLL).SDL_iconv_string("UCS-4".encode(), "UTF-8".encode(), S.encode(), SDL_GET_DLL(SDL_DLL).SDL_strlen(s.encode()) + 1)

SDL_iconv_wchar_utf8 = lambda s: \
    SDL_GET_DLL(SDL_DLL).SDL_iconv_string("UTF-8".encode(), "WCHAR_T".encode(), s.encode(), (SDL_GET_DLL(SDL_DLL).SDL_wcslen(s.encode()) + 1) * ctypes.sizeof(ctypes.c_wchar))

SDL_FunctionPointer = ctypes.CFUNCTYPE(None)